##############################################################################
#
# Copyright Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""This is a doodle/demo for generating a graphvis model from a tree.

It assumes the convention that nodes are "services" if they have a
``providers`` subnode (not a link) and if they or one of their subnodes
symbolically links to a provides node.
"""

import sys
import zc.zk

def _get_edges(tree, path, service, edges):

    if 'providers' in tree.children:
        service = path

    if service:
        for name, value in tree.properties.iteritems():
            if name.endswith(' ->') and value.endswith('/providers'):
                edges.append((service, value[:-10]))

    for name, child in tree.children.iteritems():
        _get_edges(child, path+'/'+name, service, edges)

def get_edges(tree):
    if isinstance(tree, basestring):
        if '\n' not in tree:
            if tree == '-':
                tree = sys.stdin.read()
            else:
                tree = open(tree).read()
        tree = zc.zk.parse_tree(tree)
    edges = []
    _get_edges(tree, '', '', edges)
    return edges

def dump_edges(edges, fname=None):
    if not isinstance(edges, list):
        edges = get_edges(edges)

    if fname and fname != '-':
        f = open(fname, 'w')
    else:
        f = sys.stdout

    f.write('digraph g {\n')
    for e in edges:
        f.write('  "%s" -> "%s";\n' % e)
    f.write('}\n')

    if fname and fname != '-':
        f.close()

if __name__ == '__main__':
    dump_edges(*sys.argv[1:])




##############################################################################
#
# Copyright Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__all__ = ['notify']

def notify(event):
    """NOOP event-notification hook

    Replace this with your own handler, or, better yet, install zope.event.
    """
_noop = notify

try:
    from zope.event import notify
except ImportError:
    pass



##############################################################################
#
# Copyright (c) Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Testing support

This module provides a mock of zookeeper needed to test use of zc.zk.
It's especially useful for testing packages that build on zc.zk.

It provides setUp and tearDown functions that can be used with
doctests or with regular ```unittest`` tests.
"""
from zope.testing import setupstack
import collections
import json
import mock
import os
import random
import re
import sys
import threading
import time
import traceback
import zc.zk
import zc.thread
import zookeeper

__all__ = ['assert_', 'setUp', 'tearDown', 'testing_with_real_zookeeper']

def side_effect(mock):
    return lambda func: setattr(mock, 'side_effect', func)

def assert_(cond, mess='', error=True):
    """A simple assertion function.

    If ``error``, raise an AssertionError if the assertion fails,
    otherwise, print a message.
    """
    if not cond:
        if error:
            raise AssertionError(mess)
        else:
            print 'assertion failed: ', mess

def wait_until(func=None, timeout=9):
    import warnings
    warnings.warn("wait_until is deprecated. Use zope.testing.wait.wait",
                  DeprecationWarning, 2)

    if func():
        return
    deadline = time.time()+timeout
    while not func():
        time.sleep(.01)
        if time.time() > deadline:
            raise AssertionError('timeout')

def setup_tree(tree, connection_string, root='/test-root',
               zookeeper_node=False):
    zk = zc.zk.ZooKeeper(connection_string)
    if zk.exists(root):
        zk.delete_recursive(root)
    zk.create(root, '', zc.zk.OPEN_ACL_UNSAFE)
    zk.import_tree(tree or """
    /fooservice
      /providers
      database = '/databases/foomain'
      threads = 1
      favorite_color = 'red'
    """, root)

    if zookeeper_node:
        zk.import_tree("""
        /zookeeper
          /quota
        """, root)

    zk.close()

def testing_with_real_zookeeper():
    """Test whether we're testing with a real ZooKeeper server.

    The real connection string is returned.
    """
    return os.environ.get('TEST_ZOOKEEPER_CONNECTION')

def setUp(test, tree=None, connection_string='zookeeper.example.com:2181'):
    """Set up zookeeper emulation.

    Standard (mock) testing
    -----------------------

    The first argument is a test case object (either doctest or unittest).

    You can optionally pass:

    tree
       An initial ZooKeeper tree expressed as an import string.
       If not passed, an initial tree will be created with examples
       used in the zc.zk doctests.

    connection_string
       The connection string to use for the emulation server. This
       defaults to 'zookeeper.example.com:2181'.

    Testing with a real ZooKeeper Server
    ------------------------------------

    You can test against a real ZooKeeper server, instead of a mock by
    setting the environment variable TEST_ZOOKEEPER_CONNECTION to the
    connection string of a test server.

    The tests will create a top-level node with a random name that
    starts with 'zc.zk.testing.test-roo', and use that as the virtual
    root for your tests.  Although this is the virtual root, of the
    zookeeper tree in your tests, the presense of the node may be
    shown in your tests. In particularm ``zookeeper.create`` returns
    the path created and the string returned is real, not virtual.
    This node is cleaned up by the ``tearDown``.

    A doctest can determine if it's running with a stub ZooKeeper by
    checking whether the value of the ZooKeeper gloval variable is None.
    A regular unit test can check the ZooKeeper test attribute.
    """

    globs = setupstack.globs(test)
    faux_zookeeper = None
    real_zk = testing_with_real_zookeeper()
    if real_zk:
        test_root = '/zc.zk.testing.test-root%s' % random.randint(0, sys.maxint)
        globs['/zc.zk.testing.test-root'] = test_root
        setup_tree(tree, real_zk, test_root, True)

        orig_init = zookeeper.init

        @side_effect(
            setupstack.context_manager(test, mock.patch('zookeeper.init')))
        def init(addr, watch=None, session_timeout=1000):
            if addr != connection_string:
                return orig_init(addr, watch, session_timeout)
            else:
                return orig_init(real_zk+test_root, watch, session_timeout)

        setupstack.register(
            test, lambda : setattr(zc.zk.ZooKeeper, 'test_sleep', 0))
        zc.zk.ZooKeeper.test_sleep = .01
        time.sleep(float(os.environ.get('TEST_ZOOKEEPER_SLEEP', 0)))

    else:
        if tree:
            faux_zookeeper = ZooKeeper(connection_string, Node())
        else:
            faux_zookeeper = ZooKeeper(
                connection_string,
                Node(
                    fooservice = Node(
                        json.dumps(dict(
                            database = "/databases/foomain",
                            threads = 1,
                            favorite_color= "red",
                            )),
                        providers = Node()
                        ),
                    zookeeper = Node('', quota=Node()),
                    ),
                )
        for name in ZooKeeper.__dict__:
            if name[0] == '_':
                continue
            m = setupstack.context_manager(test, mock.patch('zookeeper.'+name))
            m.side_effect = getattr(faux_zookeeper, name)

        if tree:
            zk = zc.zk.ZooKeeper(connection_string)
            zk.import_tree(tree)
            zk.close()

    globs['wait_until'] = wait_until # BBB
    globs['ZooKeeper'] = faux_zookeeper
    globs.setdefault('assert_', assert_)

def tearDown(test):
    """The matching tearDown for setUp.

    The single argument is the test case passed to setUp.
    """
    setupstack.tearDown(test)
    real_zk = testing_with_real_zookeeper()
    if real_zk:
        zk = zc.zk.ZooKeeper(real_zk)
        root = setupstack.globs(test)['/zc.zk.testing.test-root']
        if zk.exists(root):
            zk.delete_recursive(root)
        zk.close()


class Session:

    def __init__(self, zk, handle, watch=None, session_timeout=None):
        self.zk = zk
        self.handle = handle
        self.nodes = set() # ephemeral nodes
        self.add = self.nodes.add
        self.remove = self.nodes.remove
        self.watch = watch
        self.state = zookeeper.CONNECTING_STATE
        self.session_timeout = session_timeout

    def connect(self):
        self.newstate(zookeeper.CONNECTED_STATE)

    def disconnect(self):
        self.newstate(zookeeper.CONNECTING_STATE)

    def expire(self):
        self.zk._clear_session(self)
        self.newstate(zookeeper.EXPIRED_SESSION_STATE)

    def newstate(self, state):
        self.state = state
        if self.watch is not None:
            self.watch(self.handle, zookeeper.SESSION_EVENT, state, '')
        self.zk._session_event(self.handle, state)

    def check(self):
        if self.state == zookeeper.CONNECTING_STATE:
            raise zookeeper.ConnectionLossException()
        elif self.state == zookeeper.EXPIRED_SESSION_STATE:
            raise zookeeper.SessionExpiredException()
        elif self.state != zookeeper.CONNECTED_STATE:
            raise AssertionError('Invalid state')

exception_codes = {
    zookeeper.ApiErrorException: zookeeper.APIERROR,
    zookeeper.AuthFailedException: zookeeper.AUTHFAILED,
    zookeeper.BadArgumentsException: zookeeper.BADARGUMENTS,
    zookeeper.BadVersionException: zookeeper.BADVERSION,
    zookeeper.ClosingException: zookeeper.CLOSING,
    zookeeper.ConnectionLossException: zookeeper.CONNECTIONLOSS,
    zookeeper.DataInconsistencyException: zookeeper.DATAINCONSISTENCY,
    zookeeper.InvalidACLException: zookeeper.INVALIDACL,
    zookeeper.InvalidCallbackException: zookeeper.INVALIDCALLBACK,
    zookeeper.InvalidStateException: zookeeper.INVALIDSTATE,
    zookeeper.MarshallingErrorException: zookeeper.MARSHALLINGERROR,
    zookeeper.NoAuthException: zookeeper.NOAUTH,
    zookeeper.NoChildrenForEphemeralsException:
    zookeeper.NOCHILDRENFOREPHEMERALS,
    zookeeper.NoNodeException: zookeeper.NONODE,
    zookeeper.NodeExistsException: zookeeper.NODEEXISTS,
    zookeeper.NotEmptyException: zookeeper.NOTEMPTY,
    zookeeper.NothingException: zookeeper.NOTHING,
    zookeeper.OperationTimeoutException: zookeeper.OPERATIONTIMEOUT,
    zookeeper.RuntimeInconsistencyException: zookeeper.RUNTIMEINCONSISTENCY,
    zookeeper.SessionExpiredException: zookeeper.SESSIONEXPIRED,
    zookeeper.SessionMovedException: zookeeper.SESSIONMOVED,
    zookeeper.SystemErrorException: zookeeper.SYSTEMERROR,
    zookeeper.UnimplementedException: zookeeper.UNIMPLEMENTED,
}

badpath = re.compile(r'(^([^/]|$))|(/\.\.?(/|$))|(./$)').search

class ZooKeeper:

    def __init__(self, connection_string, tree):
        self.connection_strings = set([connection_string])
        self.root = tree
        self.sessions = {}
        self.lock = threading.RLock()
        self.failed = {}
        self.sequence_number = 0
        self.exists_watchers = collections.defaultdict(tuple)

    def init(self, addr, watch=None, session_timeout=4000):
        with self.lock:
            handle = 0
            while handle in self.sessions:
                handle += 1
            self.sessions[handle] = Session(
                self, handle, watch, session_timeout)
            if addr in self.connection_strings:
                self.sessions[handle].connect()
            else:
                self.failed.setdefault(addr, set()).add(handle)
            return handle

    def _allow_connection(self, connection_string):
        self.connection_strings.add(connection_string)
        for handle in self.failed.pop(connection_string, ()):
            if handle in self.sessions:
                self.sessions[handle].connect()

    def _check_handle(self, handle, checkstate=True):
        try:
            session = self.sessions[handle]
        except KeyError:
            raise zookeeper.ZooKeeperException('handle out of range')
        if checkstate:
            session.check()
        return session

    def _traverse(self, path):
        """This is used by a bunch of the methods.

        We'll test som edge cases here.

        We error on bad paths:

        >>> zk = zc.zk.ZK('zookeeper.example.com:2181')

        >>> zk.exists('')
        Traceback (most recent call last):
        ...
        BadArgumentsException: bad argument
        >>> zk.exists('xxx')
        Traceback (most recent call last):
        ...
        BadArgumentsException: bad argument
        >>> zk.exists('..')
        Traceback (most recent call last):
        ...
        BadArgumentsException: bad argument
        >>> zk.exists('.')
        Traceback (most recent call last):
        ...
        BadArgumentsException: bad argument

        >>> zk.get('')
        Traceback (most recent call last):
        ...
        BadArgumentsException: bad argument
        >>> zk.get('xxx')
        Traceback (most recent call last):
        ...
        BadArgumentsException: bad argument
        >>> zk.get('..')
        Traceback (most recent call last):
        ...
        BadArgumentsException: bad argument
        >>> zk.get('.')
        Traceback (most recent call last):
        ...
        BadArgumentsException: bad argument


        """
        if badpath(path):
            raise zookeeper.BadArgumentsException('bad argument')
        node = self.root
        for name in path.split('/')[1:]:
            if not name:
                continue
            try:
                node = node.children[name]
            except KeyError:
                raise zookeeper.NoNodeException('no node')

        return node

    def _session_event(self, handle, state):
        with self.lock:
            for path, watchers in self.exists_watchers.items():
                for h, w in watchers:
                    if h == handle:
                        w(h, zookeeper.SESSION_EVENT, state, '')
            self.root.session_event(handle, state)


    def _clear_session(self, session):
        """
        Test: don't sweat ephemeral nodes that were already deleted

        >>> zk = zc.zk.ZK('zookeeper.example.com:2181')
        >>> zk.register_server('/fooservice/providers', 'a:b')

        >>> zk2 = zc.zk.ZK('zookeeper.example.com:2181')
        >>> zk2.delete_recursive('/fooservice', force=True)
        >>> zk2.close()

        >>> zk.close()
        """
        handle = session.handle
        with self.lock:
            self.root.clear_watchers(handle)
            for path in self.exists_watchers:
                self.exists_watchers[path] = tuple(
                    (h, w) for (h, w) in self.exists_watchers[path]
                    if h != handle
                    )
            for path in list(session.nodes):
                try:
                    self._delete(session.handle, path, clear=True)
                except zookeeper.NoNodeException:
                    pass # deleted in another session, perhaps

    def _doasync(self, completion, handle, nreturn, func, *args):
        if completion is None:
            return func(*args)

        if isinstance(nreturn, int):
            nerror = nreturn
        else:
            nreturn, nerror = nreturn

        @zc.thread.Thread
        def doasync():
            try:
                # print 'doasync', func, args
                with self.lock:
                    status = 0
                    try:
                        r = func(*args)
                    except Exception, v:
                        status = exception_codes.get(v.__class__, -1)
                        r = (None, ) * nerror
                    if not isinstance(r, tuple):
                        if nreturn == 1:
                            r = (r, )
                        else:
                            r = ()
                    completion(*((handle, status) + r))
            except:
                traceback.print_exc(file=sys.stdout)

        return 0

    def close(self, handle):
        with self.lock:
            self._clear_session(self._check_handle(handle, False))
            self.sessions.pop(handle).disconnect()

    def state(self, handle):
        with self.lock:
            return self._check_handle(handle, False).state

    def create(self, handle, path, data, acl, flags=0):
        with self.lock:
            self._check_handle(handle)
            base, name = path.rsplit('/', 1)
            if flags & zookeeper.SEQUENCE:
                self.sequence_number += 1
                name += "%.10d" % self.sequence_number
                path = base + '/' + name
            if base.endswith('/'):
                raise zookeeper.BadArgumentsException('bad arguments')
            node = self._traverse(base or '/')
            for p in node.acl:
                if not (p['perms'] & zookeeper.PERM_CREATE):
                    raise zookeeper.NoAuthException('not authenticated')
            if name in node.children:
                raise zookeeper.NodeExistsException()
            node.children[name] = newnode = Node(data)
            newnode.acl = acl
            newnode.flags = flags
            node.children_changed(handle, zookeeper.CONNECTED_STATE, base)

            for h, w in self.exists_watchers.pop(path, ()):
                w(h, zookeeper.CREATED_EVENT, zookeeper.CONNECTED_STATE, path)

            if flags & zookeeper.EPHEMERAL:
                self.sessions[handle].add(path)
            return path

    def acreate(self, handle, path, data, acl, flags=0, completion=None):
        return self._doasync(completion, handle, 1,
                            self.create, handle, path, data, acl, flags)

    def _delete(self, handle, path, version=-1, clear=False):
        node = self._traverse(path)
        if version != -1 and node.version != version:
            raise zookeeper.BadVersionException('bad version')
        if node.children:
            raise zookeeper.NotEmptyException('not empty')
        base, name = path.rsplit('/', 1)
        bnode = self._traverse(base or '/')
        if not clear:
            for p in bnode.acl:
                if not (p['perms'] & zookeeper.PERM_DELETE):
                    raise zookeeper.NoAuthException('not authenticated', path)
        del bnode.children[name]
        node.deleted(handle, zookeeper.CONNECTED_STATE, path)
        bnode.children_changed(handle, zookeeper.CONNECTED_STATE, base)
        if path in self.sessions[handle].nodes:
            self.sessions[handle].remove(path)

    def delete(self, handle, path, version=-1):
        with self.lock:
            self._check_handle(handle)
            self._delete(handle, path, version)
        return 0

    def adelete(self, handle, path, version=-1, completion=None):
        return self._doasync(completion, handle, 0,
                             self.delete, handle, path, version)

    def exists(self, handle, path, watch=None):
        """Test whether a node exists:

        >>> zk = zc.zk.ZK('zookeeper.example.com:2181')
        >>> zk.exists('/test_exists')

        We can set watches:

        >>> def watch(*args):
        ...     print args

        >>> zk.exists('/test_exists', watch)
        >>> _ = zk.create('/test_exists', '', zc.zk.OPEN_ACL_UNSAFE)
        (0, 1, 3, '/test_exists')

        When a node exists, exists retirnes it's meta data, which is
        the same as the second result from get:

        >>> zk.exists('/test_exists') == zk.get('/test_exists')[1]
        True

        We can set watches on nodes that exist, too:

        >>> zk.exists('/test_exists', watch) == zk.get('/test_exists')[1]
        True

        >>> _ = zk.delete('/test_exists')
        (0, 2, 3, '/test_exists')

        Watches are one-time:

        >>> _ = zk.create('/test_exists', '', zc.zk.OPEN_ACL_UNSAFE)
        >>> _ = zk.delete('/test_exists')

        >>> zk.close()
        """
        with self.lock:
            self._check_handle(handle)
            try:
                node = self._traverse(path)
                if watch:
                    node.exists_watchers += ((handle, watch), )
                return node.meta()
            except zookeeper.NoNodeException:
                if watch:
                    self.exists_watchers[path] += ((handle, watch), )
                return None

    def aexists(self, handle, path, watch=None, completion=None):
        return self._doasync(completion, handle, 1,
                             self.exists, handle, path, watch)

    def get_children(self, handle, path, watch=None):
        with self.lock:
            self._check_handle(handle)
            node = self._traverse(path)
            for p in node.acl:
                if not (p['perms'] & zookeeper.PERM_READ):
                    raise zookeeper.NoAuthException('not authenticated')
            if watch:
                node.child_watchers += ((handle, watch), )
            return list(node.children)

    def aget_children(self, handle, path, watch=None, completion=None):
        return self._doasync(completion, handle, 1,
                             self.get_children, handle, path, watch)

    def get(self, handle, path, watch=None):
        with self.lock:
            self._check_handle(handle)
            node = self._traverse(path)
            for p in node.acl:
                if not (p['perms'] & zookeeper.PERM_READ):
                    raise zookeeper.NoAuthException('not authenticated')
            if watch:
                node.watchers += ((handle, watch), )
            return node.data, node.meta()

    def aget(self, handle, path, watch=None, completion=None):
        return self._doasync(completion, handle, 2,
                             self.get, handle, path, watch)

    def recv_timeout(self, handle):
        with self.lock:
            return self._check_handle(handle, False).session_timeout

    def set(self, handle, path, data, version=-1, async=False):
        with self.lock:
            self._check_handle(handle)
            node = self._traverse(path)
            for p in node.acl:
                if not (p['perms'] & zookeeper.PERM_WRITE):
                    raise zookeeper.NoAuthException('not authenticated')
            if version != -1 and node.version != version:
                raise zookeeper.BadVersionException('bad version')
            node.data = data
            node.changed(handle, zookeeper.CONNECTED_STATE, path)
            if async:
                return node.meta()
            else:
                return 0

    def set_watcher(self, handle, watch):
        with self.lock:
            self._check_handle(handle).watch = watch

    def aset(self, handle, path, data, version=-1, completion=None):
        return self._doasync(completion, handle, 1,
                             self.set, handle, path, data, version, True)

    def get_acl(self, handle, path):
        with self.lock:
            self._check_handle(handle)
            node = self._traverse(path)
            return node.meta(), node.acl

    def aget_acl(self, handle, path, completion=None):
        return self._doasync(completion, handle,
                             self.get_acl, handle, path)

    def set_acl(self, handle, path, aversion, acl):
        with self.lock:
            self._check_handle(handle)
            node = self._traverse(path)
            for p in node.acl:
                if not (p['perms'] & zookeeper.PERM_ADMIN):
                    raise zookeeper.NoAuthException('not authenticated', path)
            if aversion != node.aversion:
                raise zookeeper.BadVersionException("bad version")
            node.aversion += 1
            node.acl = acl

            return 0

    def aset_acl(self, handle, path, aversion, acl, completion=None):
        return self._doasync(completion, handle, 0,
                             self.set_acl, handle, path, aversion, acl)

class Node:
    watchers = child_watchers = exists_watchers = ()
    flags = 0
    version = aversion = cversion = 0
    acl = zc.zk.OPEN_ACL_UNSAFE

    def meta(self):
        return dict(
            version = self.version,
            aversion = self.aversion,
            cversion = self.cversion,
            ctime = self.ctime,
            mtime = self.mtime,
            numChildren = len(self.children),
            dataLength = len(self.data),
            ephemeralOwner=(1 if self.flags & zookeeper.EPHEMERAL else 0),
            )

    def __init__(self, data='', **children):
        self.data = data
        self.children = children
        self.ctime = self.mtime = time.time()

    def children_changed(self, handle, state, path):
        watchers = self.child_watchers
        self.child_watchers = ()
        for h, w in watchers:
            w(h, zookeeper.CHILD_EVENT, state, path)
        self.cversion += 1

    def changed(self, handle, state, path):
        watchers = self.watchers
        self.watchers = ()
        for h, w in watchers:
            w(h, zookeeper.CHANGED_EVENT, state, path)
        self.version += 1
        self.mtime = time.time()

    def deleted(self, handle, state, path):
        watchers = self.watchers
        self.watchers = ()
        for h, w in watchers:
            w(h, zookeeper.DELETED_EVENT, state, path)
        watchers = self.exists_watchers
        self.exists_watchers = ()
        for h, w in watchers:
            w(h, zookeeper.DELETED_EVENT, state, path)
        watchers = self.child_watchers
        self.watchers = ()
        for h, w in watchers:
            w(h, zookeeper.DELETED_EVENT, state, path)

    def session_event(self, handle, state):
        for (h, w) in self.watchers:
            if h == handle:
                w(h, zookeeper.SESSION_EVENT, state, '')
        for (h, w) in self.child_watchers:
            if h == handle:
                w(h, zookeeper.SESSION_EVENT, state, '')
        for (h, w) in self.exists_watchers:
            if h == handle:
                w(h, zookeeper.SESSION_EVENT, state, '')
        for child in self.children.values():
            child.session_event(handle, state)

    def clear_watchers(self, handle):
        self.watchers = tuple(
            (h, w) for (h, w) in self.watchers
            if h != handle
            )
        self.child_watchers = tuple(
            (h, w) for (h, w) in self.child_watchers
            if h != handle
            )
        self.exists_watchers = tuple(
            (h, w) for (h, w) in self.exists_watchers
            if h != handle
            )
        for name, child in self.children.items():
            child.clear_watchers(handle)



##############################################################################
#
# Copyright Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import logging
import optparse
import sys
import zc.zk
import zookeeper

def export(args=None):
    """Usage: %prog [options] connection [path]
    """
    if args is None:
        args = sys.argv[1:]

    parser = optparse.OptionParser(export.__doc__)
    parser.add_option('-e', '--ephemeral', action='store_true')
    parser.add_option('-o', '--output')

    options, args = parser.parse_args(args)
    connection = args.pop(0)
    if args:
        [path] = args
    else:
        path = '/'

    logging.basicConfig(level=logging.WARNING)

    zk = zc.zk.ZooKeeper(connection)
    data = zk.export_tree(path, ephemeral=options.ephemeral)

    if options.output:
        with open(options.output, 'w') as f:
            f.write(data)
    else:
        print data,


def import_(args=None):
    """Usage: %prog [options] connection [import-file [path]]

    Import a tree definition from a file.

    If no import-file is provided or if the import file is -, then
    data are read from standard input.
    """

    if args is None:
        args = sys.argv[1:]

    parser = optparse.OptionParser(import_.__doc__)
    parser.add_option('-d', '--dry-run', action='store_true')
    parser.add_option('-t', '--trim', action='store_true')
    parser.add_option(
        '-p', '--permission', type='int',
        default=zookeeper.PERM_ALL,
        help='ZooKeeper permission bits as integer,'
        ' defaulting to zookeeper.PERM_ALL',
        )

    options, args = parser.parse_args(args)
    if not (1 <= len(args) <= 3):
        parser.parse_args(['-h'])

    connection = args.pop(0)
    if args:
        import_file = args.pop(0)
    else:
        import_file = '-'

    if args:
        [path] = args
    else:
        path = '/'

    logging.basicConfig(level=logging.WARNING)

    zk = zc.zk.ZooKeeper(connection)
    if import_file == '-':
        import_file = sys.stdin
    else:
        import_file = open(import_file)

    zk.import_tree(
        import_file.read(), path,
        trim=options.trim,
        dry_run=options.dry_run,
        acl=[zc.zk.world_permission(options.permission)],
        )

def validate_(args=None):
    """Usage: %prog connection [file [path]]

    Validate a tree definition from a file.

    If no file is provided or if the import file is -, then
    data are read from standard input.
    """

    if args is None:
        args = sys.argv[1:]

    parser = optparse.OptionParser(import_.__doc__)
    options, args = parser.parse_args(args)
    if len(args) != 1:
        parser.parse_args(['-h'])
    if args:
        import_file = args.pop(0)
    else:
        import_file = '-'

    if import_file == '-':
        import_file = sys.stdin
    else:
        import_file = open(import_file)

    zc.zk.parse_tree(import_file.read())



##############################################################################
#
# Copyright (c) Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from pprint import pprint
from zope.testing import setupstack
from zope.testing.wait import wait
import doctest
import json
import logging
import manuel.capture
import manuel.doctest
import manuel.testing
import mock
import os
import re
import socket
import StringIO
import sys
import threading
import time
import zc.zk
import zc.zk.testing
import zc.thread
import zookeeper
import zope.testing.loggingsupport
import zope.testing.renormalizing
import unittest

class LoggingTests(unittest.TestCase):

    def test_logging(self):
        handler = zope.testing.loggingsupport.InstalledHandler(
            'ZooKeeper')
        try:
            handle = zookeeper.init('zookeeper.example.com:2181')
            zookeeper.close(handle)
        except:
            pass

        wait(lambda : [r for r in handler.records
                       if 'environment' in r.getMessage()]
             )
        handler.clear()

        # Test that the filter for the "Exceeded deadline by" noise works.
        # cheat and bypass zk by writing to the pipe directly.
        os.write(zc.zk._logging_pipe[1],
                 '2012-01-06 16:45:44,572:43673(0x1004f6000):ZOO_WARN@'
                 'zookeeper_interest@1461: Exceeded deadline by 27747ms\n')
        wait(lambda : [r for r in handler.records
                       if ('Exceeded deadline by' in r.getMessage()
                           and r.levelno == logging.DEBUG)
                       ]
             )

        self.assert_(not [r for r in handler.records
                          if ('Exceeded deadline by' in r.getMessage()
                              and r.levelno == logging.WARNING)
                          ])

        handler.uninstall()

def side_effect(mock):
    return lambda func: setattr(mock, 'side_effect', func)

class zklogger(object):

    def __init__(self):
        logger = logging.getLogger('zc.zk')
        h = logging.StreamHandler(sys.stdout)
        h.setFormatter(logging.Formatter("%(levelname)s %(message)s"))
        logger.addHandler(h)
        self.h = h
        logger.setLevel(logging.DEBUG)

    def uninstall(self):
        logger = logging.getLogger('zc.zk')
        logger.removeHandler(self.h)
        logger.setLevel(logging.NOTSET)

class Tests(unittest.TestCase):

    @mock.patch('zookeeper.init')
    def setUp(self, init):
        @zc.thread.Thread
        def getzk():
            zk = zc.zk.ZooKeeper()
            return zk

        wait(lambda : init.call_args)
        (zkaddr, self.__session_watcher), kw = init.call_args
        self.assertEqual((zkaddr, kw), ('127.0.0.1:2181', {}))
        self.__session_watcher(
            0, zookeeper.SESSION_EVENT, zookeeper.CONNECTED_STATE, '')
        getzk.join(1)
        self.__zk = getzk.value
        self.assertEqual(self.__zk.handle, 0)

        self.__teardowns = []
        cm = mock.patch('zookeeper.exists')
        @side_effect(cm.__enter__())
        def exists(handle, path):
            return True

    def tearDown(self):
        while self.__teardowns:
            self.__teardowns.pop()()

    def state_side_effect(self, handle):
        self.assertEqual(handle, self.__zk.handle)
        return zookeeper.CONNECTED_STATE

    @mock.patch('zookeeper.close')
    def test_close(self, close):
        self.__zk.close()
        close.assert_called_with(0)
        self.assertEqual(self.__zk.handle, None)

    @mock.patch('zookeeper.create')
    def test_register_server(self, create):
        @side_effect(create)
        def _(handle, path_, data, acl, flags):
            self.assertEqual((handle, path_), (0, '/foo/127.0.0.1:8080'))
            self.assertEqual(json.loads(data), dict(pid=os.getpid(), a=1))
            self.assertEqual(acl, [zc.zk.world_permission()])
            self.assertEqual(flags, zookeeper.EPHEMERAL)

        self.__zk.register_server('/foo', ('127.0.0.1', 8080), a=1)
        self.__zk.register_server('/foo', '127.0.0.1:8080', a=1)

    @mock.patch('netifaces.ifaddresses')
    @mock.patch('netifaces.interfaces')
    @mock.patch('zookeeper.create')
    def test_register_server_blank(self, create, interfaces, ifaddresses):
        addrs = {
            'eth0': {2: [{'addr': '192.168.24.60',
                          'broadcast': '192.168.24.255',
                          'netmask': '255.255.255.0'}],
                     10: [{'addr': 'fe80::21c:c0ff:fe1a:d12%eth0',
                           'netmask': 'ffff:ffff:ffff:ffff::'}],
                     17: [{'addr': '00:1c:c0:1a:0d:12',
                           'broadcast': 'ff:ff:ff:ff:ff:ff'}]},
            'foo': {2: [{'addr': '192.168.24.61',
                         'broadcast': '192.168.24.255',
                         'netmask': '255.255.255.0'}],
                    },
            'lo': {2: [{'addr': '127.0.0.1',
                        'netmask': '255.0.0.0',
                        'peer': '127.0.0.1'}],
                   10: [{'addr': '::1',
                         'netmask': 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'}],
                   17: [{'addr': '00:00:00:00:00:00',
                         'peer': '00:00:00:00:00:00'}],
                   },
            }

        @side_effect(interfaces)
        def _():
            return list(addrs)

        @side_effect(ifaddresses)
        def _(iface):
            return addrs[iface]

        @side_effect(create)
        def _(handle, path_, data, acl, flags):
            self.assertEqual(handle, 0)
            paths.append(path_)
            self.assertEqual(json.loads(data), dict(pid=os.getpid(), a=1))
            self.assertEqual(acl, [zc.zk.world_permission()])
            self.assertEqual(flags, zookeeper.EPHEMERAL)

        paths = []
        self.__zk.register_server('/foo', ('', 8080), a=1)
        self.assertEqual(sorted(paths),
                         ['/foo/192.168.24.60:8080', '/foo/192.168.24.61:8080'])

        paths = []
        self.__zk.register_server('/foo', ':8080', a=1)
        self.assertEqual(sorted(paths),
                         ['/foo/192.168.24.60:8080', '/foo/192.168.24.61:8080'])

    @mock.patch('netifaces.ifaddresses')
    @mock.patch('netifaces.interfaces')
    @mock.patch('zookeeper.create')
    def test_register_server_blank_nonet(self, create, interfaces, ifaddresses):
        addrs = {
            'lo': {2: [{'addr': '127.0.0.1',
                        'netmask': '255.0.0.0',
                        'peer': '127.0.0.1'}],
                   10: [{'addr': '::1',
                         'netmask': 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff'}],
                   17: [{'addr': '00:00:00:00:00:00',
                         'peer': '00:00:00:00:00:00'}]},
            }

        @side_effect(interfaces)
        def _():
            return list(addrs)

        @side_effect(ifaddresses)
        def _(iface):
            return addrs[iface]

        @side_effect(create)
        def _(handle, path_, data, acl, flags):
            self.assertEqual(handle, 0)
            paths.append(path_)
            self.assertEqual(json.loads(data), dict(pid=os.getpid(), a=1))
            self.assertEqual(acl, [zc.zk.world_permission()])
            self.assertEqual(flags, zookeeper.EPHEMERAL)

        paths = []
        self.__zk.register_server('/foo', ('', 8080), a=1)
        self.assertEqual(sorted(paths), ['/foo/127.0.0.1:8080'])

        paths = []
        self.__zk.register_server('/foo', ':8080', a=1)
        self.assertEqual(sorted(paths), ['/foo/127.0.0.1:8080'])

    @mock.patch('zookeeper.create')
    @mock.patch('socket.getfqdn')
    def test_register_server_blank_nonetifaces(self, getfqdn, create):

        netifaces = sys.modules['netifaces']
        try:
            sys.modules['netifaces'] = None

            @side_effect(getfqdn)
            def _():
                return 'nonexistenttestserver.zope.com'

            @side_effect(create)
            def _(handle, path_, data, acl, flags):
                self.assertEqual(
                    (handle, path_),
                    (0, '/foo/nonexistenttestserver.zope.com:8080'))
                self.assertEqual(json.loads(data), dict(pid=os.getpid(), a=1))
                self.assertEqual(acl, [zc.zk.world_permission()])
                self.assertEqual(flags, zookeeper.EPHEMERAL)

            self.__zk.register_server('/foo', ('', 8080), a=1)
            self.__zk.register_server('/foo', ':8080', a=1)
        finally:
            sys.modules['netifaces'] = netifaces

    @mock.patch('zookeeper.close')
    @mock.patch('zookeeper.init')
    @mock.patch('zookeeper.state')
    @mock.patch('zookeeper.get_children')
    def test_children(self, get_children, state, init, close):
        state.side_effect = self.state_side_effect

        path = '/test'
        @side_effect(get_children)
        def _(handle, path_, handler=None):
            if handler is not None:
                self.__handler = handler
            self.assertEqual((handle, path_), (0, path))
            return data

        # Get the data the first time
        data = []
        children = self.__zk.children(path)
        self.assertEqual(list(children), data)

        # When tree updates, children are updated
        data = ['a']
        self.__handler(0, zookeeper.CHILD_EVENT, zookeeper.CONNECTED_STATE,
                       path)
        self.assertEqual(list(children), data)

        # callbacks are called too:
        cb = mock.Mock(); children(cb)
        cb.assert_called_with(children)
        cb.reset_mock()
        self.assertEqual(len(children.callbacks), 1)
        data = ['a', 'b']
        self.__handler(0, zookeeper.CHILD_EVENT, zookeeper.CONNECTED_STATE,
                       path)
        self.assertEqual(list(children), data)
        cb.assert_called_with(children)

        # if a callback raises an exception, the exception is logged
        # and callback is discarded
        h = zope.testing.loggingsupport.Handler('zc.zk', level=logging.DEBUG)
        h.install()
        cb.side_effect = ValueError
        data = ['a']
        self.__handler(0, zookeeper.CHILD_EVENT, zookeeper.CONNECTED_STATE,
                       path)
        self.assertEqual(list(children), data)
        self.assertEqual(len(children.callbacks), 0)
        self.assertEqual(h.records[0].name, 'zc.zk')
        self.assertEqual(h.records[0].levelno, logging.ERROR)
        h.clear()

        # if a callback raises zc.zk.CancelWatch, the cancel is logged
        # and callback is discarded
        cb = mock.Mock(); children(cb)
        self.assertEqual(len(children.callbacks), 1)
        cb.side_effect = zc.zk.CancelWatch
        data = []
        self.__handler(0, zookeeper.CHILD_EVENT, zookeeper.CONNECTED_STATE,
                       path)
        self.assertEqual(list(children), data)
        self.assertEqual(len(children.callbacks), 0)
        self.assertEqual(h.records[0].name, 'zc.zk')
        self.assertEqual(h.records[0].levelno, logging.DEBUG)
        h.clear()

        h.uninstall()

        # If a session expires, it will be reestablished with watches intact.
        cb = mock.Mock(); children(cb)
        self.__session_watcher(
            0, zookeeper.SESSION_EVENT, zookeeper.EXPIRED_SESSION_STATE, "")
        close.assert_called_with(0)
        self.assertEqual(self.__zk.handle, None)
        data = ['test']
        self.__session_watcher(
            0, zookeeper.SESSION_EVENT, zookeeper.CONNECTED_STATE, "")
        self.assertEqual(list(children), data)
        cb.assert_called_with(children)


    @mock.patch('zookeeper.close')
    @mock.patch('zookeeper.init')
    @mock.patch('zookeeper.state')
    @mock.patch('zookeeper.get')
    def test_get_properties(self, get, state, init, close):
        state.side_effect = self.state_side_effect

        path = '/test'
        @side_effect(get)
        def _(handle, path_, handler=None):
            if handler is not None:
                self.__handler = handler
            self.assertEqual((handle, path_), (0, path))
            return json.dumps(data), {}

        # Get the data the first time
        data = {}
        properties = self.__zk.properties(path)
        self.assertEqual(dict(properties), data)

        # When node updates, properties are updated
        data = dict(a=1)
        self.__handler(0, zookeeper.CHANGED_EVENT, zookeeper.CONNECTED_STATE,
                     path)
        self.assertEqual(dict(properties), data)

        # callbacks are called too:
        cb = mock.Mock(); properties(cb)
        cb.assert_called_with(properties)
        cb.reset_mock()
        self.assertEqual(len(properties.callbacks), 1)
        self.__handler(0, zookeeper.CHANGED_EVENT, zookeeper.CONNECTED_STATE,
                     path)
        self.assertEqual(dict(properties), data)
        cb.assert_called_with(properties)

        # if a callback raises an exception, the exception is logged
        # and callback is discarded
        h = zope.testing.loggingsupport.Handler('zc.zk', level=logging.DEBUG)
        h.install()
        cb.side_effect = ValueError
        data = dict(a=1)
        self.__handler(0, zookeeper.CHANGED_EVENT, zookeeper.CONNECTED_STATE,
                     path)
        self.assertEqual(dict(properties), data)
        self.assertEqual(len(properties.callbacks), 0)
        self.assertEqual(h.records[0].name, 'zc.zk')
        self.assertEqual(h.records[0].levelno, logging.ERROR)
        h.clear()

        # if a callback raises zc.zk.CancelWatch, the cancel is logged
        # and callback is discarded
        cb = mock.Mock(); properties(cb)
        self.assertEqual(len(properties.callbacks), 1)
        cb.side_effect = zc.zk.CancelWatch
        data = {}
        self.__handler(0, zookeeper.CHANGED_EVENT, zookeeper.CONNECTED_STATE,
                     path)
        self.assertEqual(dict(properties), data)
        self.assertEqual(len(properties.callbacks), 0)
        self.assertEqual(h.records[0].name, 'zc.zk')
        self.assertEqual(h.records[0].levelno, logging.DEBUG)
        h.clear()

        h.uninstall()

        # If a session expires, it will be reestablished with watches intact.
        cb = mock.Mock(); properties(cb)
        self.__session_watcher(
            0, zookeeper.SESSION_EVENT, zookeeper.EXPIRED_SESSION_STATE, "")
        close.assert_called_with(0)
        self.assertEqual(self.__zk.handle, None)
        data = dict(test=1)
        self.__session_watcher(
            0, zookeeper.SESSION_EVENT, zookeeper.CONNECTED_STATE, "")
        self.assertEqual(dict(properties), data)
        cb.assert_called_with(properties)

    @mock.patch('zookeeper.state')
    @mock.patch('zookeeper.get')
    @mock.patch('zookeeper.set')
    def test_set_properties(self, set, get, state):
        state.side_effect = self.state_side_effect

        path = '/test'
        @side_effect(get)
        def _(handle, path_, handler=None):
            if handler is not None:
                self.__handler = handler
            self.assertEqual((handle, path_), (0, path))
            return json.dumps(data), {}

        data = dict(a=1)
        properties = self.__zk.properties(path)
        self.assertEqual(dict(properties), data)

        @side_effect(set)
        def _(handle, path_, data, version=-1):
            self.__set_data = json.loads(data)
            self.assertEqual((handle, path_), (0, path))

        properties.update(b=2)
        self.assertEqual(self.__set_data, dict(a=1, b=2))
        self.assertEqual(dict(properties), self.__set_data)

        properties.set(c=3)
        self.assertEqual(self.__set_data, dict(c=3))
        self.assertEqual(dict(properties), self.__set_data)

    @mock.patch('zookeeper.state')
    @mock.patch('zookeeper.get')
    @mock.patch('zookeeper.set')
    def test_special_values(self, set, get, state):
        state.side_effect = self.state_side_effect

        path = '/test'
        @side_effect(get)
        def _(handle, path_, handler=None):
            if handler is not None:
                self.__handler = handler
            self.assertEqual((handle, path_), (0, path))
            return data, {}

        data = ''
        properties = self.__zk.properties(path)
        self.assertEqual(dict(properties), {})

        data = 'xxx'
        properties = self.__zk.properties(path)
        self.assertEqual(dict(properties), dict(string_value='xxx'))

        data = '{xxx}'
        properties = self.__zk.properties(path)
        self.assertEqual(dict(properties), dict(string_value='{xxx}'))

        data = '\n{xxx}\n'
        properties = self.__zk.properties(path)
        self.assertEqual(dict(properties), dict(string_value='\n{xxx}\n'))

        @side_effect(set)
        def _(handle, path_, data, version=-1):
            self.__set_data = data
            self.assertEqual((handle, path_), (0, path))

        properties.set(b=2)
        self.assertEqual(self.__set_data, '{"b":2}')
        properties.set()
        self.assertEqual(self.__set_data, '')
        properties.set(string_value='xxx')
        self.assertEqual(self.__set_data, 'xxx')

def test_children():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> _ = zk.create('/test', '', zc.zk.OPEN_ACL_UNSAFE)
    >>> children = zk.children('/test')
    >>> sorted(children)
    []

    >>> len(children)
    0
    >>> bool(children)
    False

    >>> def create(path):
    ...     zk.create(path, '', zc.zk.OPEN_ACL_UNSAFE)
    >>> create('/test/a')
    >>> sorted(children)
    ['a']

    >>> len(children)
    1
    >>> bool(children)
    True

We can register callbacks:

    >>> @children
    ... def cb(c):
    ...     print 'good', sorted(c)
    good ['a']

When we register a callback, it gets called immediately with a children object.

    >>> create('/test/b')
    good ['a', 'b']
    >>> sorted(children)
    ['a', 'b']

If a callback raises an error immediately, it isn't saved:

    >>> @children
    ... def bad(c):
    ...     raise ValueError
    Traceback (most recent call last):
    ...
    ValueError

    >>> create('/test/c')
    good ['a', 'b', 'c']

    >>> len(children)
    3
    >>> bool(children)
    True

If a callback raises an error later, it is logged and the callback is
cancelled:

    >>> logger = zklogger()

    >>> badnow = False
    >>> @children
    ... def bad(c):
    ...     assert c is children
    ...     print 'bad later', sorted(c)
    ...     if badnow:
    ...         raise ValueError
    bad later ['a', 'b', 'c']

    >>> _ = zk.delete('/test/c')
    good ['a', 'b']
    bad later ['a', 'b']

    >>> badnow = True
    >>> _ = zk.delete('/test/b') # doctest: +ELLIPSIS
    good ['a']
    bad later ['a']
    ERROR watch(zc.zk.Children(0, /test), <function bad at ...>)
    Traceback (most recent call last):
    ...
    ValueError

    >>> _ = zk.delete('/test/a')
    good []

A callback can also cancel itself by raising CancelWatch:

    >>> cancelnow = False
    >>> @children
    ... def cancel(c):
    ...     assert c is children
    ...     print 'cancel later', sorted(c)
    ...     if cancelnow:
    ...         raise zc.zk.CancelWatch
    cancel later []

    >>> create('/test/a')
    good ['a']
    cancel later ['a']

    >>> cancelnow = True
    >>> create('/test/b') # doctest: +ELLIPSIS
    good ['a', 'b']
    cancel later ['a', 'b']
    DEBUG cancelled watch(zc.zk.Children(0, /test), <function cancel at ...>)

    >>> logger.uninstall()

    >>> zk.close()

    """

def test_handler_cleanup():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> _ = zk.create('/test', '', zc.zk.OPEN_ACL_UNSAFE)

Children:

    >>> children = zk.children('/test')
    >>> len(zk.watches)
    1
    >>> del children
    >>> len(zk.watches)
    0

    >>> children = zk.children('/test')
    >>> @children
    ... def kids(c):
    ...     print c
    zc.zk.Children(0, /test)
    >>> len(zk.watches)
    1
    >>> del children
    >>> len(zk.watches)
    1
    >>> del kids
    >>> len(zk.watches)
    0

    >>> @zk.children('/test')
    ... def kids(c):
    ...     print c
    zc.zk.Children(0, /test)

    >>> len(zk.watches)
    1
    >>> del kids
    >>> len(zk.watches)
    0

Properties:

    >>> properties = zk.properties('/test')
    >>> len(zk.watches)
    1
    >>> del properties
    >>> len(zk.watches)
    0

    >>> properties = zk.properties('/test')
    >>> @properties
    ... def props(c):
    ...     print c
    zc.zk.Properties(0, /test)
    >>> len(zk.watches)
    1
    >>> del properties
    >>> len(zk.watches)
    1
    >>> del props
    >>> len(zk.watches)
    0

    >>> @zk.properties('/test')
    ... def props(c):
    ...     print c
    zc.zk.Properties(0, /test)

    >>> len(zk.watches)
    1
    >>> del props
    >>> len(zk.watches)
    0

    >>> zk.close()

    """

def test_deleted_node_with_watchers():
    """

Set up some handlers.

    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> _ = zk.create('/test', '{"a": 1}', zc.zk.OPEN_ACL_UNSAFE)

    >>> children = zk.children('/test')
    >>> @children
    ... def _(arg):
    ...     print 1, list(arg)
    1 []

    >>> @children
    ... def _(arg=None):
    ...     print 2, arg
    2 zc.zk.Children(0, /test)

    >>> _ = zk.create('/test/a', '', zc.zk.OPEN_ACL_UNSAFE)
    1 ['a']
    2 zc.zk.Children(0, /test)

    >>> _ = zk.delete('/test/a')
    1 []
    2 zc.zk.Children(0, /test)

    >>> properties = zk.properties('/test')
    >>> @properties
    ... def _(arg):
    ...     print 3, dict(arg)
    3 {u'a': 1}

    >>> @properties
    ... def _(arg=None):
    ...     print 4, arg
    4 zc.zk.Properties(0, /test)

    >>> zk.set('/test', '{"b": 2}')
    3 {u'b': 2}
    4 zc.zk.Properties(0, /test)
    0

Hack data into the child watcher to verify it's cleared:

    >>> children.data = 'data'

Now delete the node.  The handlers that accept no arguments will be called:

    >>> _ = zk.delete('/test')
    4 None
    2 None

Note that the handlers that accept 0 arguments were called.

And the data are cleared:

    >>> list(children), list(properties)
    ([], [])

    >>> zk.close()
    """


def resilient_import():
    """
We can use vatious spacing in properties and links:

    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.import_tree('''
    ... /test
    ...   a=1
    ...   b      =1
    ...   c=      1
    ...   ad->/x
    ...   af          ->/x
    ...   ae->         /x
    ... ''')

    >>> print zk.export_tree('/test'),
    /test
      a = 1
      b = 1
      c = 1
      ad -> /x
      ae -> /x
      af -> /x

When an expression is messed up, we get sane errors:

    >>> zk.import_tree('''
    ... /test
    ...   a= 1+
    ... ''') # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    ValueError: Error unexpected EOF while parsing (<string>, line 1)
    in expression: '1+' in line 3

    >>> zk.import_tree('''
    ... /test
    ...   a ->
    ... ''') # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    ValueError: (3, 'a ->', 'Bad link format')

    >>> zk.close()
    """

def import_dry_run():
    """

    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.import_tree('''
    ... /test
    ...   a=1
    ...   b = 2
    ...   /c1
    ...     /c12
    ...   /c2
    ...   ae->         /x
    ...   ad->         /y
    ... ''')

    >>> zk.import_tree('''
    ... /test
    ...   a=2
    ...   /c1
    ...     /c12
    ...       a = 1
    ...       b -> /b
    ...       /c123
    ...   ae->         /z
    ... ''', dry_run=True)
    /test a change from 1 to 2
    /test remove link ad -> /y
    /test ae link change from /x to /z
    /test remove property b = 2
    extra path not trimmed: /test/c2
    /test/c1/c12 add property a = 1
    /test/c1/c12 add link b -> /b
    add /test/c1/c12/c123

    >>> print zk.export_tree('/test'),
    /test
      a = 1
      b = 2
      ad -> /y
      ae -> /x
      /c1
        /c12
      /c2

    >>> zk.close()
    """

def property_set_and_update_variations():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> data = zk.properties('/fooservice')
    >>> @data
    ... def _(data):
    ...     pprint(dict(data), width=70)
    {u'database': u'/databases/foomain',
     u'favorite_color': u'red',
     u'threads': 1}

    >>> data.set(dict(x=1))
    {u'x': 1}
    >>> data.set(dict(x=1), x=2, y=3)
    {u'x': 2, u'y': 3}
    >>> data.set(z=1)
    {u'z': 1}
    >>> data.update(a=1)
    {u'a': 1, u'z': 1}
    >>> data.update(dict(b=1), a=2)
    {u'a': 2, u'b': 1, u'z': 1}
    >>> data.update(dict(c=1))
    {u'a': 2, u'b': 1, u'c': 1, u'z': 1}
    >>> data.update(dict(d=1), d=2)
    {u'a': 2, u'b': 1, u'c': 1, u'd': 2, u'z': 1}

    >>> zk.close()
    """

def test_resolve():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.import_tree('''
    ... /top
    ...  /a
    ...    top -> /top
    ...    loop -> /top/a/b/loop
    ...    /b
    ...      top -> /top
    ...      loop -> /top/a/loop
    ...      /c
    ...        top -> /top
    ...        /d
    ...          name = 'd'
    ...          /e
    ...          top -> /top
    ...          loop -> /top/a/b/c/d/loop
    ...
    ... ''')


    >>> zk.resolve('/top/a/b/c/d')
    '/top/a/b/c/d'

    >>> zk.resolve('/top/a/top/a/b/top/a/b/c/top/a/b/c/d')
    u'/top/a/b/c/d'

    >>> sorted(zk.properties('/top/a/top/a/b/top/a/b/c/top/a/b/c/d').items())
    [(u'loop ->', u'/top/a/b/c/d/loop'), (u'name', u'd'), (u'top ->', u'/top')]

    >>> zk.register_server('/top/a/top/a/b/top/a/b/c/top/a/b/c/d', 'addr')
    >>> sorted(map(str, zk.children('/top/a/top/a/b/top/a/b/c/top/a/b/c/d')))
    ['addr', 'e']

    >>> zk.resolve('/top/a/top/a/b/top/x')
    Traceback (most recent call last):
    ...
    NoNodeException: /top/a/top/a/b/top/x

    >>> zk.resolve('/top/a/b/c/d/loop')
    Traceback (most recent call last):
    ...
    LinkLoop: ('/top/a/b/c/d/loop', u'/top/a/b/c/d/loop')

    >>> zk.resolve('/top/a/loop/b/c/d')
    Traceback (most recent call last):
    ...
    LinkLoop: ('/top/a/loop', u'/top/a/b/loop', u'/top/a/loop')

    >>> zk.resolve('/top/a/b/c/d/./../..')
    '/top/a/b'

    >>> zk.close()
    """

def test_ln_target_w_trailing_slash():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.ln('/databases/main', '/fooservice/')
    >>> pprint(zk.get_properties('/fooservice'))
    {u' ->': u'/databases/main',
     u'database': u'/databases/foomain',
     u'favorite_color': u'red',
     u'threads': 1}

    >>> zk.close()
    """

def test_export_top_w_name():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> print zk.export_tree('/', name='top'),
    /top
      /fooservice
        database = u'/databases/foomain'
        favorite_color = u'red'
        threads = 1
        /providers

    >>> zk.close()
    """

def test_set():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.get('/')[0]
    ''
    >>> zk.set('/', 'a'); zk.get('/')[0]
    0
    'a'

    >>> zk.set('/', 'b', 0)
    Traceback (most recent call last):
    ...
    BadVersionException: bad version

    >>> zk.set('/', 'b'); zk.get('/')[0]
    0
    'b'

    >>> r = zk.aset('/', 'c', -1, check_async()); event.wait(1)
    ... # doctest: +ELLIPSIS
    async callback got (...
    >>> r
    0

    >>> zk.get('/')[0]
    'c'

    >>> r = zk.aset('/', 'd', 0,
    ...             check_async(expected_status=zookeeper.BADVERSION)
    ...             ); event.wait(1)
    async callback got (None,)

    >>> r
    0

    >>> r = zk.aset('/', 'd', 3, check_async()); event.wait(1)
    ... # doctest: +ELLIPSIS
    async callback got (...
    >>> r
    0

>>> time.sleep(1)

    >>> zk.get('/')[0]
    'd'

    >>> zk.close()
    """

def test_delete():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> _ = zk.create('/test', '', zc.zk.OPEN_ACL_UNSAFE)

Synchronous variations:

    >>> _ = zk.create('/test/a', '', zc.zk.OPEN_ACL_UNSAFE)
    >>> _ = zk.set('/test/a', '1')
    >>> _ = zk.set('/test/a', '2')
    >>> zk.delete('/test/a', 0)
    Traceback (most recent call last):
    ...
    BadVersionException: bad version

    >>> zk.get_children('/test')
    ['a']
    >>> _ = zk.delete('/test/a', 2)
    >>> zk.get_children('/test')
    []

    >>> _ = zk.create('/test/a', '', zc.zk.OPEN_ACL_UNSAFE)
    >>> _ = zk.set('/test/a', '1')
    >>> _ = zk.set('/test/a', '2')
    >>> _ = zk.delete('/test/a', -1)
    >>> zk.get_children('/test')
    []

    >>> _ = zk.create('/test/a', '', zc.zk.OPEN_ACL_UNSAFE)
    >>> _ = zk.set('/test/a', '1')
    >>> _ = zk.set('/test/a', '2')
    >>> _ = zk.delete('/test/a')
    >>> zk.get_children('/test')
    []


Asynchronous variations:

    >>> _ = zk.create('/test/a', '', zc.zk.OPEN_ACL_UNSAFE)
    >>> _ = zk.set('/test/a', '1')
    >>> _ = zk.set('/test/a', '2')
    >>> r = zk.adelete('/test/a', 0,
    ...             check_async(expected_status=zookeeper.BADVERSION)
    ...             ); event.wait(1)
    async callback got ()
    >>> r
    0

    >>> zk.get_children('/test')
    ['a']
    >>> r = zk.adelete('/test/a', 2, check_async()); event.wait(1)
    async callback got ()
    >>> r, zk.get_children('/test')
    (0, [])

    >>> _ = zk.create('/test/a', '', zc.zk.OPEN_ACL_UNSAFE)
    >>> _ = zk.set('/test/a', '1')
    >>> _ = zk.set('/test/a', '2')
    >>> r = zk.adelete('/test/a', -1, check_async()); event.wait(1)
    async callback got ()
    >>> r, zk.get_children('/test')
    (0, [])

    >>> _ = zk.create('/test/a', '', zc.zk.OPEN_ACL_UNSAFE)
    >>> _ = zk.set('/test/a', '1')
    >>> _ = zk.set('/test/a', '2')
    >>> r = zk.adelete('/test/a', completion=check_async()); event.wait(1)
    async callback got ()
    >>> r, zk.get_children('/test')
    (0, [])

    >>> zk.close()
    """

def test_set_acl():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> _ = zk.create('/test', '', zc.zk.OPEN_ACL_UNSAFE)
    >>> zk.get_acl('/test')[1] == zc.zk.OPEN_ACL_UNSAFE
    True
    >>> zk.set_acl('/test', 0, [zc.zk.world_permission(17)])
    0
    >>> zk.get_acl('/test')[1] == [zc.zk.world_permission(17)]
    True

    >>> zk.set_acl('/test', 0, [zc.zk.world_permission(17)])
    Traceback (most recent call last):
    ...
    BadVersionException: bad version

    >>> zk.set_acl('/test', 1, [zc.zk.world_permission(18)])
    0
    >>> zk.get_acl('/test')[1] == [zc.zk.world_permission(18)]
    True

    >>> r = zk.aset_acl('/test', 0, [zc.zk.world_permission(19)],
    ...                 check_async(expected_status=zookeeper.BADVERSION)
    ...                 ); event.wait(1)
    async callback got ()
    >>> r
    0

    >>> r = zk.aset_acl('/test', 2, [zc.zk.world_permission(19)],
    ...                 check_async()); event.wait(1)
    async callback got ()
    >>> r
    0
    >>> zk.get_acl('/test')[1] == [zc.zk.world_permission(19)]
    True

    >>> zk.close()
    """

def test_server_registeration_event():
    """
    >>> import sys, zc.zk.event, zope.event
    >>> zc.zk.event.notify is zope.event.notify
    True
    >>> sys.modules['zope.event'] = None
    >>> _ = reload(zc.zk.event)
    >>> zc.zk.event.notify is zope.event.notify
    False
    >>> zc.zk.event.notify is zc.zk.event._noop
    True
    >>> def notify(e):
    ...     print e
    ...     e.properties['test'] = 1
    >>> zc.zk.event.notify = notify

    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.register_server('/fooservice/providers', '1.2.3.4:5678')
    RegisteringServer('1.2.3.4:5678', '/fooservice/providers', {'pid': 1793})

    >>> zk.print_tree('/fooservice/providers')
    /providers
      /1.2.3.4:5678
        pid = 1793
        test = 1

    >>> zk.close()

    >>> sys.modules['zope.event'] = zope.event
    >>> _ = reload(zc.zk.event)
    >>> zc.zk.event.notify is zope.event.notify
    True
    """

def connection_edge_cases():
    """
We can pass a session timeout, and it will be passed to ZooKeeper:

    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181', 4242)
    >>> zk.recv_timeout()
    4242
    >>> zk.close()

Connecting to an invalid address caused a FailedConnect to be raised:

    >>> zc.zk.ZooKeeper('192.0.2.42:2181')
    Traceback (most recent call last):
    ...
    FailedConnect: 192.0.2.42:2181

    """

def register_server_at_root():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.register_server('/', 'a:b')
    >>> zk.print_tree() # doctest: +ELLIPSIS
    /a:b
      pid = 2318
    /fooservice
    ...
    >>> zk.close()
    """

relative_property_links_data = """

/a
  /b
    x => c x
    xx => ./c x
    x2 => .. x
    x3 => ../c x
    x22 => ./.. x
    x33 => ./../c x
    x333 => .././c x
    /c
      x=1
  /c
    x = 3
  x = 2

"""
def relative_property_links():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.import_tree(relative_property_links_data)
    >>> p = zk.properties('/a/b')
    >>> p['x']
    1
    >>> p['xx']
    1
    >>> p['x2']
    2
    >>> p['x22']
    2
    >>> p['x3']
    3
    >>> p['x33']
    3
    >>> p['x333']
    3
    >>> zk.close()
    """

def property_links_expand_callbacks_to_linked_nodes():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.import_tree('''
    ... /a
    ...   /b
    ...     x => c
    ...     xx => ../c
    ...     /c
    ...       x = 1
    ...   /c
    ...     x = 2
    ... ''')

    >>> ab = zk.properties('/a/b')

    >>> @ab
    ... def _(properties):
    ...     print 'updated'
    updated

    >>> ac = zk.properties('/a/c')
    >>> ac.update(x=3)
    updated

    >>> ab.update(xx=2)
    updated

    >>> ac.update(x=4)

    >>> zk.close()
    """

def bad_links_are_reported_and_prevent_updates():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> properties = zk.properties('/fooservice')

    >>> properties.update({'c =>': '/a/b/c d'}, a=1, b=2, d=3, e=4)
    ... # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    ValueError: ('Bad property link', 'c =>', '/a/b/c d',
    NoNodeException('/a/b/c',))
    >>> pprint(dict(properties), width=70)
    {u'database': u'/databases/foomain',
     u'favorite_color': u'red',
     u'threads': 1}

    >>> properties.update({'c =>': ''}, a=1, b=2, d=3, e=4)
    ... # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    ValueError: ('Bad property link', 'c =>', '',
    ValueError('Bad link data',))
    >>> pprint(dict(properties), width=70)
    {u'database': u'/databases/foomain',
     u'favorite_color': u'red',
     u'threads': 1}

    >>> properties.update({'c =>': '/fooservice x'}, a=1, b=2, d=3, e=4)
    Traceback (most recent call last):
    ...
    ValueError: ('Bad property link', 'c =>', '/fooservice x', KeyError('x',))
    >>> pprint(dict(properties), width=70)
    {u'database': u'/databases/foomain',
     u'favorite_color': u'red',
     u'threads': 1}

    >>> properties.update({'c =>': '/fooservice threads x'}, a=1, b=2, d=3, e=4)
    ... # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    ValueError: ('Bad property link', 'c =>', '/fooservice threads x',
    ValueError('Bad link data',))
    >>> pprint(dict(properties), width=70)
    {u'database': u'/databases/foomain',
     u'favorite_color': u'red',
     u'threads': 1}

    >>> properties.set({'c =>': '/a/b/c d'}, a=1, b=2, d=3, e=4)
    ... # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    ValueError: ('Bad property link', 'c =>', '/a/b/c d',
    NoNodeException('/a/b/c',))
    >>> pprint(dict(properties), width=70)
    {u'database': u'/databases/foomain',
     u'favorite_color': u'red',
     u'threads': 1}

    >>> properties.set({'c =>': ''}, a=1, b=2, d=3, e=4)
    Traceback (most recent call last):
    ...
    ValueError: ('Bad property link', 'c =>', '', ValueError('Bad link data',))
    >>> pprint(dict(properties), width=70)
    {u'database': u'/databases/foomain',
     u'favorite_color': u'red',
     u'threads': 1}

    >>> zk.close()
    """

def contains_w_property_link():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> properties = zk.properties('/fooservice/providers')
    >>> properties.update({'c =>': '.. threads'})
    >>> 'c' in properties
    True

    >>> zk.close()
    """

def property_getitem_error_handling():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> _ = zk.set('/fooservice/providers', json.dumps({
    ... 'a =>': '/a/b',
    ... 'b =>': '/fooservice threads x',
    ... 'c =>': '',
    ... }))
    >>> properties = zk.properties('/fooservice/providers')
    >>> properties['a']
    Traceback (most recent call last):
    ...
    BadPropertyLink: (NoNodeException(u'/a/b',), "in 'a =>': u'/a/b'")
    >>> properties['b'] # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    BadPropertyLink: (ValueError('Invalid property link',),
    "in 'b =>': u'/fooservice threads x'")
    >>> properties['c']
    Traceback (most recent call last):
    ...
    BadPropertyLink: (IndexError('pop from empty list',), "in 'c =>': u''")

    >>> zk.close()
    """

def property_link_loops():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.import_tree('''
    ... /a
    ...   x => ../b x
    ... /b
    ...   x => ../a x
    ... ''')
    >>> properties = zk.properties('/a')
    >>> properties['x'] # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
    ...
    BadPropertyLink:
    (BadPropertyLink(BadPropertyLink(LinkLoop((u'/b', u'/a', u'/b'),),
    "in u'x =>': u'../b x'"), "in u'x =>': u'../a x'"), "in 'x =>': u'../b x'")

    >>> zk.close()
    """

def deleting_linked_nodes():
    """
    Links are reestablished after deleting a linked node.

    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.import_tree('''
    ... /a
    ...   /b
    ...     x => ../c
    ...   /c
    ...     x = 1
    ... ''')

    >>> ab = zk.properties('/a/b')

    >>> @ab
    ... def _(properties):
    ...     print 'updated'
    updated

    >>> ab['x']
    1

    >>> zk.import_tree('''
    ... /d
    ...   x = 2
    ... ''')

    >>> _ = zk.set('/a', '{"c ->": "/d"}')
    >>> _ = zk.delete('/a/c')
    updated
    >>> ab['x']
    2

    >>> zk.close()
    """

def delete_recursive_dry_run():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.delete_recursive('/fooservice', dry_run=True)
    would delete /fooservice/providers.
    would delete /fooservice.

    >>> zk.print_tree()
    /fooservice
      database = u'/databases/foomain'
      favorite_color = u'red'
      threads = 1
      /providers

    >>> zk.close()
    """

def delete_recursive_force():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.register_server('/fooservice/providers', 'a:b')

    >>> zk.delete_recursive('/fooservice', dry_run=True, force=True)
    would delete /fooservice/providers/a:b.
    would delete /fooservice/providers.
    would delete /fooservice.

    >>> zk.print_tree()
    /fooservice
      database = u'/databases/foomain'
      favorite_color = u'red'
      threads = 1
      /providers
        /a:b
          pid = 29093

    >>> zk.delete_recursive('/fooservice', force=True)

    >>> zk.print_tree()
    <BLANKLINE>

    >>> zk.close()
    """

def is_ephemeral():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.register_server('/fooservice/providers', 'a:b')
    >>> zk.is_ephemeral('/fooservice')
    False
    >>> zk.is_ephemeral('/fooservice/providers')
    False
    >>> zk.is_ephemeral('/fooservice/providers/a:b')
    True
    >>> zk.close()
    """

def create_recursive():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.create_recursive('/fooservice/a/b/c', '', zc.zk.OPEN_ACL_UNSAFE)
    >>> acl = [dict(perms=zookeeper.PERM_CREATE |
    ...        zookeeper.PERM_READ |
    ...        zookeeper.PERM_DELETE,
    ...        scheme='world', id='anyone')]
    >>> zk.create_recursive('/a/b/c', '{"z": 1}', acl)

    >>> zk.print_tree()
    /a
      z = 1
      /b
        z = 1
        /c
          z = 1
    /fooservice
      database = u'/databases/foomain'
      favorite_color = u'red'
      threads = 1
      /a
        /b
          /c
      /providers

    >>> for path in zk.walk('/fooservice/a'):
    ...     if zk.get_acl(path)[1] != zc.zk.OPEN_ACL_UNSAFE:
    ...         print 'oops'

    >>> for path in zk.walk('/a'):
    ...     if zk.get_acl(path)[1] != acl:
    ...         print 'oops'

    >>> zk.close()
    """

def property_links_edge_cases():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.import_tree('''
    ... /app
    ...   threads => ..
    ...   color = 'red'
    ...   database -> /databases/foo
    ... ''', '/fooservice')
    >>> zk.print_tree()
    /fooservice
      database = u'/databases/foomain'
      favorite_color = u'red'
      threads = 1
      /app
        color = u'red'
        threads => = u'..'
        database -> /databases/foo
      /providers

    >>> properties = zk.properties('/fooservice/app')
    >>> sorted(properties)
    [u'color', u'database ->', u'threads']

    >>> sorted(properties.keys())
    [u'color', u'database ->', u'threads']

    >>> sorted(properties.values())
    [1, u'/databases/foo', u'red']

    >>> sorted(properties.items())
    [(u'color', u'red'), (u'database ->', u'/databases/foo'), (u'threads', 1)]

    >>> pprint(dict(properties))
    {u'color': u'red', u'database ->': u'/databases/foo', u'threads': 1}

    >>> zk.close()
    """

def no_spam_when_not_trimming_ephemeral_nodes():
    """
    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> zk.print_tree()
    /fooservice
      database = u'/databases/foomain'
      favorite_color = u'red'
      threads = 1
      /providers

    >>> zk.register_server('/fooservice/providers', 'a:a')
    >>> zk.print_tree()
    /fooservice
      database = u'/databases/foomain'
      favorite_color = u'red'
      threads = 1
      /providers
        /a:a
          pid = 1234

    >>> zk.import_tree('''
    ... /fooservice
    ...   database = u'/databases/foomain'
    ...   favorite_color = u'red'
    ...   threads = 1
    ...   /providers
    ... ''', trim=True)

    >>> zk.close()
    """

event = threading.Event()
def check_async(show=True, expected_status=0):
    event.clear()
    def check(handle, status, *args):
        if show:
            print 'async callback got', args
        event.set()
        zc.zk.testing.assert_(
            status==expected_status,
            "Bad cb status %s" % status,
            error=False)
    return check

def setUpEphemeral_node_recovery_on_session_reestablishment(test):
    zc.zk.testing.setUp(test)
    test.globs['check_async'] = check_async
    test.globs['event'] = event

def setUpREADME(test):
    zc.zk.testing.setUp(test)
    @side_effect(setupstack.context_manager(test, mock.patch('socket.getfqdn')))
    def getfqdn():
        return 'socket.getfqdn'

def disconnectiontestsSetup(test):
    zc.zk.testing.setUp(test)
    setupstack.register(
        test, setattr, zc.zk.ZooKeeper, 'initial_connection_wait',
        zc.zk.ZooKeeper.initial_connection_wait)
    zc.zk.ZooKeeper.initial_connection_wait = .1

checker = zope.testing.renormalizing.RENormalizing([
    (re.compile('pid = \d+'), 'pid = 9999'),
    (re.compile("{'pid': \d+}"), 'pid = 9999'),
    (re.compile('/zc\.zk\.testing\.test-root\d+'), ''),
    (re.compile(r'2 None\n4 None'), '4 None\n2 None'),
    ])

def test_suite():
    suite = unittest.TestSuite((
        manuel.testing.TestSuite(
            manuel.doctest.Manuel(
                checker=checker, optionflags=doctest.NORMALIZE_WHITESPACE) +
            manuel.capture.Manuel(),
            'README.txt',
            setUp=setUpREADME, tearDown=zc.zk.testing.tearDown,
            ),
        doctest.DocTestSuite(
            setUp=zc.zk.testing.setUp, tearDown=zc.zk.testing.tearDown,
            checker=checker,
            ),
        doctest.DocFileSuite(
            'monitor.test',
            checker = zope.testing.renormalizing.RENormalizing([
                (re.compile(':\d+'), ':9999'),
                ])
            ),
        ))
    if not zc.zk.testing.testing_with_real_zookeeper():
        suite.addTest(unittest.makeSuite(Tests))
        suite.addTest(unittest.makeSuite(LoggingTests))
        suite.addTest(doctest.DocFileSuite(
            'ephemeral_node_recovery_on_session_reestablishment.test',
            setUp=setUpEphemeral_node_recovery_on_session_reestablishment,
            tearDown=zc.zk.testing.tearDown,
            checker=checker,
            ))
        suite.addTest(doctest.DocTestSuite(
            'zc.zk.disconnectiontests',
            setUp=disconnectiontestsSetup, tearDown=zc.zk.testing.tearDown,
            checker=checker,
            ))
        suite.addTest(doctest.DocTestSuite(
            'zc.zk.testing',
            setUp=zc.zk.testing.setUp, tearDown=zc.zk.testing.tearDown,
            checker=checker,
            )),

    return suite



##############################################################################
#
# Copyright Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

# This module has tests that involve zookeeper being disconnected.

# Their in a separate module to avoid testing with a real zookeeper
# server, which we can't control (or at least don't want to work hard
# enough to control).

from pprint import pprint
from zope.testing.wait import wait
import zc.zk
import zookeeper
import zope.testing.loggingsupport

def wait_for_zookeeper():
    """
    Normally, zc.zk.ZooKeeper raises an exception if it can't connect
    to ZooKeeper in a short time [#initial_connection_wait]_.  Some
    applications might want to wait, so zc.zk.ZooKeeper accepts a wait
    parameter that causes it to wait for a connection.

    >>> zk = None
    >>> import zc.thread

    >>> handler = zope.testing.loggingsupport.InstalledHandler('zc.zk')

    >>> @zc.thread.Thread
    ... def connect():
    ...     global zk
    ...     zk = zc.zk.ZooKeeper('Invalid', wait=True)

    We'll wait a while while it tries in vane to connect:

    >>> wait((lambda : zk is not None), 4)
    Traceback (most recent call last):
    ...
    TimeOutWaitingFor: <lambda>

    >>> print handler # doctest: +ELLIPSIS
    zc.zk CRITICAL
      Can't connect to ZooKeeper at 'Invalid'
    zc.zk CRITICAL
      Can't connect to ZooKeeper at 'Invalid'
    ...
    >>> handler.uninstall()

    Now, we'll make the connection possible:

    >>> ZooKeeper._allow_connection('Invalid')
    >>> wait(lambda : zk is not None)

    >>> zk.state == zookeeper.CONNECTED_STATE
    True

    Yay!

    >>> zk.close()

    .. [#initial_connection_wait] The initial wait is configurable,
       mainly for testing, via zc.zk.ZooKeeper.initial_connection_wait
    """

def session_timeout_with_child_and_data_watchers():
    """

Set up a session with some watchers:

    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')

    >>> handler = zope.testing.loggingsupport.InstalledHandler('zc.zk')

    >>> properties = zk.properties('/fooservice')
    >>> @properties
    ... def changed(a):
    ...     print 'properties changed', a is properties
    properties changed True

    >>> pprint(dict(properties), width=60)
    {u'database': u'/databases/foomain',
     u'favorite_color': u'red',
     u'threads': 1}

    >>> children = zk.children('/fooservice')
    >>> @children
    ... def changed(a):
    ...     print 'children changed', a is children
    children changed True

    >>> sorted(children)
    ['providers']

Now, we'll expire the session:

    >>> handler.clear()
    >>> ZooKeeper.sessions[zk.handle].disconnect()
    >>> ZooKeeper.sessions[zk.handle].expire()
    children changed True
    properties changed True

(Note that we got the handlers called when we reestablished the new
 session.  This is important as the data may have changed between the
 old and new session.)

Now, if we make changes, they'll be properly reflected:

    >>> _ = zk.set('/fooservice', '{"a": 1}')
    properties changed True

    >>> dict(properties)
    {u'a': 1}

    >>> zk.register_server('/fooservice', 'x')
    children changed True

    >>> sorted(children)
    ['providers', 'x']

    >>> print handler
    zc.zk INFO
      connected 0

    """

def session_events_are_ignored_by_child_and_data_watch_support():
    """Session events are send to child, data and exists watcher.

    This should have no impact on the watch support.

    Nothing should get logged. No errors, no warnings.

    >>> zk = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> handler = zope.testing.loggingsupport.InstalledHandler('zc.zk')
    >>> @zk.properties('/fooservice')
    ... def p(data):
    ...     print 'property changed'
    property changed

    >>> @zk.children('/fooservice')
    ... def c(data):
    ...     print 'children changed'
    children changed

    >>> node = ZooKeeper._traverse('/fooservice')
    >>> len(node.watchers), len(node.child_watchers)
    (1, 1)

    >>> for state in (zookeeper.CONNECTING_STATE,
    ...               zookeeper.CONNECTED_STATE,
    ...               zookeeper.EXPIRED_SESSION_STATE,
    ...     ):
    ...     for h, w in (node.watchers + node.child_watchers):
    ...         w(h, zookeeper.SESSION_EVENT, state, '')

    >>> print handler,
    >>> handler.uninstall()

    >>> len(node.watchers), len(node.child_watchers)
    (1, 1)

    >>> zk2 = zc.zk.ZooKeeper('zookeeper.example.com:2181')
    >>> p2 = zk2.properties('/fooservice')
    >>> p2.update(z=1)
    property changed
    >>> _ = zk2.create('/fooservice/xxx', '', zc.zk.OPEN_ACL_UNSAFE)
    children changed

    >>> len(node.watchers), len(node.child_watchers)
    (2, 1)

    >>> zk.close()
    >>> zk2.close()
    """



##############################################################################
#
# Copyright (c) Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import collections
import json
import logging
import os
import re
import socket
import sys
import threading
import time
import weakref
import zc.zk.event
import zc.thread
import zookeeper

logger = logging.getLogger(__name__)

_logging_pipe = os.pipe()
zookeeper.set_log_stream(os.fdopen(_logging_pipe[1], 'w'))

@zc.thread.Thread
def loggingthread():
    r, w = _logging_pipe
    log = logging.getLogger('ZooKeeper').log
    f = os.fdopen(r)
    levels = dict(ZOO_INFO = logging.INFO,
                  ZOO_WARN = logging.WARNING,
                  ZOO_ERROR = logging.ERROR,
                  ZOO_DEBUG = logging.DEBUG,
                  )
    while 1:
        line = f.readline().strip()
        if not line:
            continue
        try:
            if '@' in line:
                level, message = line.split('@', 1)
                level = levels.get(level.split(':')[-1])

                if 'Exceeded deadline by' in line and level == logging.WARNING:
                    level = logging.DEBUG

            else:
                level = None

            if level is None:
                log(logging.INFO, line)
            else:
                log(level, message)
        except Exception, v:
            logging.getLogger('ZooKeeper').exception("Logging error: %s", v)


def parse_addr(addr):
    host, port = addr.split(':')
    return host, int(port)

def encode(props):
    if len(props) == 1 and 'string_value' in props:
        return props['string_value']

    if props:
        return json.dumps(props, separators=(',',':'))
    else:
        return ''

def decode(sdata, path='?'):
    s = sdata.strip()
    if not s:
        data = {}
    elif s.startswith('{') and s.endswith('}'):
        try:
            data = json.loads(s)
        except:
            logger.exception('bad json data in node at %r', path)
            data = dict(string_value = sdata)
    else:
        data = dict(string_value = sdata)
    return data

def join(*args):
    return '/'.join(args)

def world_permission(perms=zookeeper.PERM_READ):
    return dict(perms=perms, scheme='world', id='anyone')

OPEN_ACL_UNSAFE = [world_permission(zookeeper.PERM_ALL)]
READ_ACL_UNSAFE = [world_permission()]

class CancelWatch(Exception):
    pass

class LinkLoop(Exception):
    pass

class FailedConnect(Exception):
    pass

class BadPropertyLink(Exception):
    pass

class Resolving:

    def resolve(self, path, seen=()):
        if path.endswith('/.'):
            return self.resolve(path[:-2])
        if path.endswith('/..'):
            base = self.resolve(path[:-3])
            if '/' in base:
                return base.rsplit('/', 1)[0]
            else:
                raise zookeeper.NoNodeException(path)

        try:
            if self.exists(path):
                return path
        except zookeeper.BadArgumentsException:
            if not path[:1] == '/':
                raise zookeeper.NoNodeException(path)

        if path in seen:
            seen += (path,)
            raise LinkLoop(seen)

        try:
            base, name = path.rsplit('/', 1)
            base = self.resolve(base, seen)
            newpath = base + '/' + name
            if self.exists(newpath):
                return newpath
            props = self.get_properties(base)
            newpath = props.get(name+' ->')
            if not newpath:
                raise zookeeper.NoNodeException()

            if not newpath[0] == '/':
                newpath = base + '/' + newpath

            seen += (path,)
            return self.resolve(newpath, seen)
        except zookeeper.NoNodeException:
            raise zookeeper.NoNodeException(path)


class ZooKeeper(Resolving):

    initial_connection_wait = 9.0

    def __init__(self, connection_string="127.0.0.1:2181", session_timeout=None,
                 wait=False):
        self.watches = WatchManager()
        self.ephemeral = {}
        self.handle = None

        connected = self.connected = threading.Event()
        def watch_session(handle, event_type, state, path):
            assert event_type == zookeeper.SESSION_EVENT
            assert not path
            if state == zookeeper.CONNECTED_STATE:
                if self.handle is None:
                    self.handle = handle
                    for watch in self.watches.clear():
                        self._watch(watch)
                    for path, data in self.ephemeral.items():
                        zookeeper.create(self.handle, path, data['data'],
                                         data['acl'], data['flags'])
                else:
                    assert handle == self.handle
                connected.set()
                logger.info('connected %s', handle)
            elif state == zookeeper.CONNECTING_STATE:
                connected.clear()
            elif state == zookeeper.EXPIRED_SESSION_STATE:
                connected.clear()
                if self.handle is not None:
                    zookeeper.close(self.handle)
                self.handle = None
                init()
            else:
                logger.critical('unexpected session event %s %s', handle, state)

        if session_timeout:
            init = (lambda : zookeeper.init(connection_string, watch_session,
                                            session_timeout)
                    )
        else:
            init = lambda : zookeeper.init(connection_string, watch_session)

        handle = init()
        connected.wait(self.initial_connection_wait)
        if not connected.is_set():
            if wait:
                while not connected.is_set():
                    logger.critical("Can't connect to ZooKeeper at %r",
                                    connection_string)
                    connected.wait(1)
            else:
                zookeeper.close(handle)
                raise FailedConnect(connection_string)


    def _findallipv4addrs(self, tail):
        try:
            import netifaces
        except ImportError:
            return [socket.getfqdn()+tail]

        addrs = set()
        loopaddrs = set()
        for iface in netifaces.interfaces():
            for info in netifaces.ifaddresses(iface).get(2, ()):
                addr = info.get('addr')
                if addr:
                    if addr.startswith('127.'):
                        loopaddrs.add(addr+tail)
                    else:
                        addrs.add(addr+tail)

        return addrs or loopaddrs

    def register_server(self, path, addr, acl=READ_ACL_UNSAFE, **kw):
        kw['pid'] = os.getpid()

        if not isinstance(addr, str):
            addr = '%s:%s' % tuple(addr)

        if addr[:1] == ':':
            addrs = self._findallipv4addrs(addr)
        else:
            addrs = (addr,)

        path = self.resolve(path)
        zc.zk.event.notify(RegisteringServer(addr, path, kw))
        if path != '/':
            path += '/'

        for addr in addrs:
            self.create(path + addr, encode(kw), acl, zookeeper.EPHEMERAL)

    test_sleep = 0
    def _async(self, completion, meth, *args):
        post = getattr(self, '_post_'+meth)
        if completion is None:
            result = getattr(zookeeper, meth)(self.handle, *args)
            post(*args)
            if self.test_sleep:
                time.sleep(self.test_sleep)
            return result

        def asynccb(handle, status, *cargs):
            assert handle == self.handle
            if status == 0:
                post(*args)
            completion(handle, status, *cargs)

        return getattr(zookeeper, 'a'+meth)(self.handle, *(args+(asynccb,)))

    def create(self, path, data, acl, flags=0, completion=None):
        return self._async(completion, 'create', path, data, acl, flags)
    acreate = create

    def _post_create(self, path, data, acl, flags):
        if (flags & zookeeper.EPHEMERAL) and not (flags & zookeeper.SEQUENCE):
            self.ephemeral[path] = dict(data=data, acl=acl, flags=flags)

    def delete(self, path, version=-1, completion=None):
        return self._async(completion, 'delete', path, version)
    adelete = delete

    def _post_delete(self, path, version):
        self.ephemeral.pop(path, None)

    def set(self, path, data, version=-1, completion=None):
        return self._async(completion, 'set', path, data, version)
    aset = set2 = set

    def _post_set(self, path, data, version):
        if path in self.ephemeral:
            self.ephemeral[path]['data'] = data

    def set_acl(self, path, version, acl, completion=None):
        return self._async(completion, 'set_acl', path, version, acl)
    aset_acl = set_acl

    def _post_set_acl(self, path, version, acl):
        if path in self.ephemeral:
            self.ephemeral[path]['acl'] = acl

    def _watch(self, watch):
        event_type = watch.event_type
        watch.real_path = real_path = self.resolve(watch.path)
        key = event_type, real_path
        if self.watches.add(key, watch):
            try:
                self._watchkey(key)
            except zookeeper.ConnectionLossException:
                # We lost a race here. We got disconnected between
                # when we resolved the watch path and the time we set
                # the watch. This is very unlikely.
                watches = set(self.watches.pop(key))
                for w in watches:
                    w._deleted()
                if watch in watches:
                    watches.remove(watch)
                if watches:
                    # OMG, how unlucky can we be?
                    # someone added a watch between the time we added
                    # the key and failed to add the watch in zookeeper.
                    logger.critical('lost watches %r', watches)
                raise
        else:
            # We already had a watch for the key.  We need to pass this one
            # it's data.
            zkfunc = getattr(zookeeper, self.__zkfuncs[event_type])
            watch._notify(zkfunc(self.handle, real_path))

    __zkfuncs = {
        zookeeper.CHANGED_EVENT: 'get',
        zookeeper.CHILD_EVENT: 'get_children',
        }
    def _watchkey(self, key):
        event_type, real_path = key
        zkfunc = getattr(zookeeper, self.__zkfuncs[event_type])

        def handler(h, t, state, p, reraise=False):

            if t == zookeeper.SESSION_EVENT:
                return

            if state != zookeeper.CONNECTED_STATE:
                # This can happen if we get disconnected or a session expires.
                # When we reconnect, we should restablish the watchers.
                logger.warning(
                    "Node watcher event %r with non-connected state, %r",
                    t, state)
                return

            try:
                assert h == self.handle
                assert state == zookeeper.CONNECTED_STATE
                assert p == real_path
                if key not in self.watches:
                    return

                if t == zookeeper.DELETED_EVENT:
                    self._rewatch(key)
                else:
                    assert t == event_type
                    try:
                        v = zkfunc(self.handle, real_path, handler)
                    except zookeeper.NoNodeException:
                        self._rewatch(key)
                    else:
                        for watch in self.watches.watches(key):
                            watch._notify(v)
            except:
                logger.exception("%s(%s) handler failed",
                                 self.__zkfuncs[event_type], real_path)
                if reraise:
                    raise

        handler(self.handle, event_type, self.state, real_path, True)

    def _rewatch(self, key):
        event_type = key[0]
        for watch in self.watches.pop(key):
            try:
                real_path = self.resolve(watch.path)
            except (zookeeper.NoNodeException, LinkLoop):
                logger.exception("%s path went away", watch)
                watch._deleted()
            else:
                self._watch(watch)

    def children(self, path):
        return Children(self, path)

    def create_recursive(self, path, data, acl):
        if self.exists(path):
            return
        base, name = path.rsplit('/', 1)
        if base:
            self.create_recursive(base, data, acl)
        if not self.exists(path):
            self.create(path, data, acl)

    def get_properties(self, path):
        return decode(self.get(path)[0])

    def properties(self, path):
        return Properties(self, path)

    def import_tree(self, text, path='/', trim=False, acl=OPEN_ACL_UNSAFE,
                    dry_run=False):
        while path.endswith('/'):
            path = path[:-1] # Mainly to deal w root: /
        self._import_tree(path, parse_tree(text), acl, trim, dry_run, True)

    def _import_tree(self, path, node, acl, trim, dry_run, top=False):
        if not top:
            new_children = set(node.children)
            for name in sorted(self.get_children(path)):
                if name in new_children:
                    continue
                cpath = join(path, name)
                if trim:
                    self.delete_recursive(cpath, dry_run,
                                          ignore_if_ephemeral=True)
                else:
                    print 'extra path not trimmed:', cpath

        for name, child in sorted(node.children.iteritems()):
            cpath = path + '/' + name
            data = encode(child.properties)
            if self.exists(cpath):
                if dry_run:
                    new = child.properties
                    old = self.get_properties(cpath)
                    old = decode(self.get(cpath)[0])
                    for n, v in sorted(old.items()):
                        if n not in new:
                            if n.endswith(' ->'):
                                print '%s remove link %s %s' % (cpath, n, v)
                            else:
                                print '%s remove property %s = %s' % (
                                    cpath, n, v)
                        elif new[n] != v:
                            if n.endswith(' ->'):
                                print '%s %s link change from %s to %s' % (
                                    cpath, n[:-3], v, new[n])
                            else:
                                print '%s %s change from %s to %s' % (
                                    cpath, n, v, new[n])
                    for n, v in sorted(new.items()):
                        if n not in old:
                            if n.endswith(' ->'):
                                print '%s add link %s %s' % (cpath, n, v)
                            else:
                                print '%s add property %s = %s' % (
                                    cpath, n, v)
                else:
                    self.set(cpath, data)
                    meta, oldacl = self.get_acl(cpath)
                    if acl != oldacl:
                        self.set_acl(cpath, meta['aversion'], acl)
            else:
                if dry_run:
                    print 'add', cpath
                    continue
                else:
                    self.create(cpath, data, acl)
            self._import_tree(cpath, child, acl, trim, dry_run)

    def delete_recursive(self, path, dry_run=False, force=False,
                         ignore_if_ephemeral=False):
        self._delete_recursive(path, dry_run, force, ignore_if_ephemeral)

    def _delete_recursive(self, path, dry_run, force,
                          ignore_if_ephemeral=False):
        ephemeral_child = None
        for name in sorted(self.get_children(path)):
            ephemeral_child = (
                self._delete_recursive(join(path, name), dry_run, force) or
                ephemeral_child
                )

        if ephemeral_child:
            print "%s not deleted due to ephemeral descendent." % path
            return ephemeral_child

        ephemeral = self.is_ephemeral(path) and not force
        if ephemeral and ignore_if_ephemeral:
            return
        if dry_run:
            if ephemeral:
                print "wouldn't delete %s because it's ephemeral." % path
            else:
                print "would delete %s." % path
        else:
            if ephemeral:
                print "Not deleting %s because it's ephemeral." % path
            else:
                logger.info('deleting %s', path)
                self.delete(path)
        return ephemeral

    def is_ephemeral(self, path):
        return bool(self.get(path)[1]['ephemeralOwner'])

    def export_tree(self, path='/', ephemeral=False, name=None):
        output = []
        out = output.append

        def export_tree(path, indent, name=None):
            children = self.get_children(path)
            if path == '/':
                path = ''
                if 'zookeeper' in children:
                    children.remove('zookeeper')
                if name is not None:
                    out(indent + '/' + name)
                    indent += '  '
            else:
                data, meta = self.get(path)
                if meta['ephemeralOwner'] and not ephemeral:
                    return
                if name is None:
                    name = path.rsplit('/', 1)[1]
                properties = decode(data)
                type_ = properties.pop('type', None)
                if type_:
                    name += ' : '+type_
                out(indent + '/' + name)
                indent += '  '
                links = []
                for i in sorted(properties.iteritems()):
                    if i[0].endswith(' ->'):
                        links.append(i)
                    else:
                        out(indent+"%s = %r" % i)
                for i in links:
                    out(indent+"%s %s" % i)

            for name in sorted(children):
                export_tree(path+'/'+name, indent)

        export_tree(path, '', name)
        return '\n'.join(output)+'\n'

    def print_tree(self, path='/'):
        print self.export_tree(path, True),

    def _set(self, path, data):
        return self.set(path, data)

    def ln(self, target, source):
        base, name = source.rsplit('/', 1)
        if target[-1] == '/':
            target += name
        properties = self.get_properties(base)
        properties[name+' ->'] = target
        self._set(base, encode(properties))

    def close(self):
        zookeeper.close(self.handle)
        self.handle = None

    @property
    def state(self):
        if self.handle is None:
            return zookeeper.CONNECTING_STATE
        return zookeeper.state(self.handle)

    def walk(self, path='/', ephemeral=True, children=False):
        try:
            if not ephemeral and self.get(path)[1]['ephemeralOwner']:
                return

            _children = sorted(self.get_children(path))
            if children:
                yield path, _children
            else:
                yield path
        except zookeeper.NoNodeException:
            return

        for name in _children:
            if path != '/':
                name = '/'+name
            for p in self.walk(path+name):
                yield p


def _make_method(name):
    return (lambda self, *a, **kw:
            getattr(zookeeper, name)(self.handle, *a, **kw))

for name in (
    'add_auth', 'aexists', 'aget', 'aget_acl',
    'aget_children', 'async', 'client_id',
    'exists', 'get', 'get_acl',
    'get_children', 'is_unrecoverable', 'recv_timeout',
    ):
    setattr(ZooKeeper, name, _make_method(name))

del _make_method

class WatchManager:
    # Manage {key -> w{watches}} in a thread-safe manner.
    # (And also provide a hard set to allow nodeinfos w callbacks
    #  to keep themselves around.)

    def __init__(self):
        self.data = {}
        self.lock = threading.Lock()

        def _remove(ref, selfref=weakref.ref(self)):
            self = selfref()
            if self is None:
                return
            key = ref.key
            with self.lock:
                refs = self.data.get(key, ())
                if ref in refs:
                    refs.remove(ref)
                    if not refs:
                        del self.data[key]

        self._remove = _remove

    def __len__(self):
        with self.lock:
            return sum(
                len([r for r in refs if r() is not None])
                for refs in self.data.itervalues()
                )

    def __contains__(self, key):
        with self.lock:
            return key in self.data

    def add(self, key, value):
        ref = weakref.KeyedRef(value, self._remove, key)
        newkey = False
        with self.lock:
            try:
                refs = self.data[key]
            except KeyError:
                self.data[key] = refs = set()
                newkey = True
            refs.add(ref)
        return newkey

    def pop(self, key):
        with self.lock:
            watches = [ref() for ref in self.data.pop(key, ())]

        for watch in watches:
            if watch is not None:
                yield watch

    def watches(self, key):
        with self.lock:
            watches = [ref() for ref in self.data.get(key, ())]

        for watch in watches:
            if watch is not None:
                yield watch

    def clear(self):
        # Clear data and return an iterator on the old values
        with self.lock:
            old = self.data
            self.data = {}

        for refs in old.itervalues():
            for ref in refs:
                v = ref()
                if v is not None:
                    yield v

ZK = ZooKeeper

class NodeInfo:

    def __init__(self, session, path):
        self.session = session
        self.path = path
        self.callbacks = []
        session._watch(self)

    def setData(self, data):
        self.data = data

    deleted = False
    def _deleted(self):
        self.deleted = True
        self.data = {}
        for callback in self.callbacks:
            try:
                callback()
            except TypeError:
                pass
            except:
                logger.exception('Error %r calling %r', self, callback)

    def __repr__(self):
        return "%s%s.%s(%s, %s)" % (
            self.deleted and 'DELETED: ' or '',
            self.__class__.__module__, self.__class__.__name__,
            self.session.handle, self.path)

    def _notify(self, data):
        if data is not None:
            self.setData(data)
        for callback in list(self.callbacks):
            try:
                callback(self)
            except Exception, v:
                self.callbacks.remove(callback)
                if isinstance(v, CancelWatch):
                    logger.debug("cancelled watch(%r, %r)", self, callback)
                else:
                    logger.exception("watch(%r, %r)", self, callback)

    def __call__(self, func):
        func(self)
        self.callbacks.append(func)
        return self

    def __iter__(self):
        return iter(self.data)

class Children(NodeInfo):

    event_type = zookeeper.CHILD_EVENT

    def __len__(self):
        return len(self.data)

class Properties(NodeInfo, collections.Mapping):

    event_type = zookeeper.CHANGED_EVENT

    def __init__(self, *args):
        self._linked_properties = {}
        NodeInfo.__init__(self, *args)

    def _setData(self, data, handle_errors=False):
        # Save a mapping as our data.
        # Set up watchers for any property links.
        _linked_properties = {}
        for name in data:
            if name.endswith(' =>') and name[:-3] not in data:
                link = data[name].strip().split()
                try:
                    if not (1 <= len(link) <= 2):
                        raise ValueError('Bad link data')
                    path = link.pop(0)
                    if path[0] != '/':
                        path = self.path + '/' + path
                    path = self.session.resolve(path)
                    properties = self._setup_link(path, _linked_properties)
                    properties[link and link[0] or name[:-3]]
                except Exception, v:
                    if handle_errors:
                        logger.exception(
                            'Bad property link %r %r', name, data[name])
                    else:
                        raise ValueError("Bad property link",
                                         name, data[name], v)

        # Only after processing all links, do we update instance attrs:
        self.data = data
        self._linked_properties = _linked_properties

    def _setup_link(self, path, _linked_properties=None):
        if _linked_properties is None:
            _linked_properties = self._linked_properties

        props = _linked_properties.get(path, self)
        if props is not self:
            return props

        props = self.session.properties(path)

        _linked_properties[path] = props

        def notify(properties=None):
            if properties is None:
                # A node we were watching was deleted.  We shuld try to
                # re-resolve it. This doesn't happen often, let's just reset
                # everything.
                self._setData(self.data)
            elif self._linked_properties.get(path) is properties:
                self._notify(None)
            else:
                # We must not care about it anymore.
                raise CancelWatch()

        props.callbacks.append(notify)
        return props

    def setData(self, data):
        # Called by upstream watchers.
        sdata, self.meta_data = data
        self._setData(decode(sdata, self.path), True)

    def __getitem__(self, key, seen=()):
        try:
            return self.data[key]
        except KeyError:
            link = self.data.get(key + ' =>', self)
            if link is self:
                raise
            try:
                data = link.split()
                if len(data) > 2:
                    raise ValueError('Invalid property link')
                path = data.pop(0)
                if not path[0] == '/':
                    path = self.path + '/' + path

                path = self.session.resolve(path)
                if path in seen:
                    raise LinkLoop(seen+(path,))
                seen += (path,)
                properties = self._linked_properties.get(path)
                if properties is None:
                    properties = self._setup_link(path)
                name = data and data[0] or key
                return properties.__getitem__(name, seen)
            except Exception, v:
                raise BadPropertyLink(
                    v, 'in %r: %r' %
                    (key + ' =>', self.data[key + ' =>'])
                    )

    def __iter__(self):
        for key in self.data:
            if key.endswith(' =>'):
                key = key[:-3]
            yield key

    def __len__(self):
        return len(self.data)

    def __contains__(self, key):
        return key in self.data or (key + ' =>') in self.data

    def copy(self):
        return self.data.copy()

    def _set(self, data):
        self._setData(data)
        self.session._set(self.path, encode(data))

    def set(self, data=None, **properties):
        data = data and dict(data) or {}
        data.update(properties)
        self._set(data)

    def update(self, data=None, **properties):
        d = self.data.copy()
        if data:
            d.update(data)
        d.update(properties)
        self._set(d)

    def __hash__(self):
        # Gaaaa, collections.Mapping
        return hash(id(self))

_text_is_node = re.compile(
    r'/(?P<name>\S+)'
    '(\s*:\s*(?P<type>\S.*))?'
    '$').match
_text_is_property = re.compile(
    r'(?P<name>\S+)'
    '\s*=\s*'
    '(?P<expr>\S.*)'
    '$'
    ).match
_text_is_link = re.compile(
    r'(?P<name>\S+)'
    '\s*->\s*'
    '(?P<target>\S+)'
    '$'
    ).match
_text_is_plink = re.compile(
    r'(?P<name>\S+)'
    '\s*=>\s*'
    '(?P<target>\S+(\s+\S+)?)'
    '(\s+(?P<pname>/\S+))?'
    '$'
    ).match

class ParseNode:

    def __init__(self, name='', properties=None, **children):
        self.name = name
        self.properties = properties or {}
        self.children = children
        for name, child in children.iteritems():
            child.name = name

def parse_tree(text, node_class=ParseNode):
    root = node_class()
    indents = [(-1, root)] # sorted [(indent, node)]
    lineno = 0
    for line in text.split('\n'):
        lineno += 1
        line = line.rstrip()
        if not line:
            continue
        stripped = line.strip()
        if stripped[0] == '#':
            continue
        indent = len(line) - len(stripped)

        data = None

        m = _text_is_plink(stripped)
        if m:
            data = (m.group('name') + ' =>'), m.group('target')

        if data is None:
            m = _text_is_property(stripped)
            if m:
                expr = m.group('expr')
                try:
                    data = eval(expr, {})
                except Exception, v:
                    raise ValueError(
                        "Error %s in expression: %r in line %s" %
                        (v, expr, lineno))
                data = m.group('name'), data

        if data is None:
            m = _text_is_link(stripped)
            if m:
                data = (m.group('name') + ' ->'), m.group('target')

        if data is None:
            m = _text_is_node(stripped)
            if m:
                data = node_class(m.group('name'))
                if m.group('type'):
                    data.properties['type'] = m.group('type')

        if data is None:
            if '->' in stripped:
                raise ValueError(lineno, stripped, "Bad link format")
            else:
                raise ValueError(lineno, stripped, "Unrecognized data")

        if indent > indents[-1][0]:
            if not isinstance(indents[-1][1], node_class):
                raise ValueError(
                    lineno, line,
                    "Can't indent under properties")
            indents.append((indent, data))
        else:
            while indent < indents[-1][0]:
                indents.pop()

            if indent > indents[-1][0]:
                raise ValueError(lineno, data, "Invalid indentation")

        if isinstance(data, node_class):
            children = indents[-2][1].children
            if data.name in children:
                raise ValueError(lineno, data, 'duplicate node')
            children[data.name] = data
            indents[-1] = indent, data
        else:
            if indents[-2][1] is root:
                raise ValueError("Can't above imported nodes.")
            properties = indents[-2][1].properties
            name, value = data
            if name in properties:
                raise ValueError(lineno, data, 'duplicate property')
            properties[name] = value

    return root

class RegisteringServer:
    """Event emitted while a server is being registered.

    Attributes:

    name
      The server name (node name)
    name
      The service path (node parent path)
    properties
      A dictionary of properties to be saved on the ephemeral node.

      Typeically, subscribers will add properties.
    """

    def __init__(self, name, path, properties):
        self.name = name
        self.path = path
        self.properties = properties

    def __repr__(self):
        return "RegisteringServer(%r, %r, %r)" % (
            self.name, self.path, self.properties)



##############################################################################
#
# Copyright (c) Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
name, version = 'zc.zk', '0'

install_requires = ['setuptools', 'zc.thread']
extras_require = dict(
    test=['zope.testing >= 4.1.0', 'zc-zookeeper-static', 'mock', 'manuel',
          'zope.event', 'netifaces', 'zope.component', 'zc.monitor'],
    static=['zc-zookeeper-static'],
    )

entry_points = """
[console_scripts]
zookeeper_export = zc.zk.scripts:export
zookeeper_import = zc.zk.scripts:import_
zookeeper_validate = zc.zk.scripts:validate_
"""

from setuptools import setup
import os
readme = open(os.path.join('src', 'zc', 'zk', 'README.txt')).read()

setup(
    author = 'Jim Fulton',
    author_email = 'jim@zope.com',
    license = 'ZPL 2.1',

    name = name, version = version,
    long_description=readme,
    description = readme.strip().split('\n')[1],
    packages = [name.split('.')[0], name],
    namespace_packages = [name.split('.')[0]],
    package_dir = {'': 'src'},
    install_requires = install_requires,
    zip_safe = False,
    entry_points=entry_points,
    include_package_data = True,
    extras_require = extras_require,
    tests_require = extras_require['test'],
    test_suite = name+'.tests.test_suite',
    )



##############################################################################
#
# Copyright (c) 2011 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import json
import logging
import sys

logger = logging.getLogger(__name__)

_servers = []

def notify(event):
    _servers.append(dict(address=event.name, path=event.path,
                         **event.properties))

def servers(connection, path=None):
    if path is None:
        connection.write(json.dumps(_servers) + '\n')
    else:
        connection.write(
            ' '.join([s['address'] for s in _servers if s['path'] == path])
             + '\n')

def _connect(addr):
    import re
    import socket

    if re.search(':\d+$', addr):
        host, port = addr.rsplit(':')
        addr = host, int(port)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    else:
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

    sock.connect(addr)
    return sock

def check(args=None):
    if args is None:
        args = sys.argv[1:]

    [addr, path] = args

    try:
        sock = _connect(addr)

        sock.sendall('servers %s\n' % path)
        f = sock.makefile()
        addr = f.readline()

        f.close()
        sock.close()

        _connect(addr).close()
    except Exception:
        logger.debug("Failed check", exc_info=True)
        sys.exit(1)

def get_addr(args=None):
    if args is None:
        args = sys.argv[1:]


    [addr, path] = args

    sock = _connect(addr)
    f = sock.makefile()
    sock.sendall('servers %s\n' % path)
    print f.readline(),
